###############################################################
# Name:			 Preferences.py
# Purpose:	 Hold program preferences
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2010-01-31
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

from . import i18n
import datetime

from .SongFormat import *
from .decorators import StandardVerseNumbers
from .Transpose import *


i18n.register('songpress.Preferences')

def get_update_frequencies():
	return {
		0: _("Never"),
		7: _("Week"),
		14: _("Two weeks"),
		30: _("Month"),
		60: _("Two months"),
	}
	
class Preferences(object):
	"""
		Available preferences

		* format
		* decoratorFormat
		* decorator
		* chorusLabel (None or string)
		* labelVerses
		* editorFace
		* editorSize
		* fontFace
		* defaultNotation
		* notations
		* defaultNotation
		* autoAdjustSpuriousLines
		* autoAdjustTab2Chordpro
		* autoAdjustEasyKey
		* locale
		* updateFrequency (days, or 0 for never)
		* ignoredUpdates
		* updateUrl
		* updateLastCheck
		* Set/GetEasyChordsGroup
		* GetEasyChords
	"""
	def __init__(self):
		object.__init__(self)
		self.config = wx.Config.Get()
		self.format = SongFormat()
		self.decoratorFormat = StandardVerseNumbers.Format(self.format, _("Chorus"))
		self.decorator = StandardVerseNumbers.Decorator(self.decoratorFormat)
		self.notations = [enNotation, itNotation, deNotation, tradDeNotation, frNotation, ptNotation]
		self.easyChordsGroup = {}
		self.Load()

	def SetFont(self, font, showChords=None):
		self.fontFace = font
		self.format.face = font
		self.format.comment.face = font
		self.format.chord.face = font
		self.format.chorus.face = font
		self.format.chorus.chord.face = font
		self.format.chorus.comment.face = font
		for v in self.format.verse:
			v.face = font
			v.chord.face = font
			v.comment.face = font
		self.format.title.face = font
		self.format.subtitle.face = font
		self.decoratorFormat.face = font
		self.decoratorFormat.chorus.face = font
		if showChords is not None:
			self.format.showChords = showChords

	def Load(self):
		self.notices = {}
		self.config.SetPath('/Format')
		l = self.config.Read('ChorusLabel')
		if l:
			self.chorusLabel = l
			self.decoratorFormat.SetChorusLabel(l)
		else:
			self.chorusLabel = None
		showChords = int(self.config.Read('ShowChords', "2"))
		self.config.SetPath('/Format/Font')
		face = self.config.Read('Face')
		if face:
			self.fontFace = face
		else:
			self.fontFace = "Arial"
		self.SetFont(self.fontFace, showChords)
		self.config.SetPath('/Format/Style')
		labelVerses = self.config.Read('LabelVerses')
		if labelVerses:
			self.labelVerses = bool(int(labelVerses))
		else:
			self.labelVerses = True
		self.config.SetPath('/Editor')
		self.editorFace = self.config.Read('Face')
		self.editorSize = self.config.Read('Size')
		if not self.editorFace:
			self.editorFace = wx.SystemSettings().GetFont(wx.SYS_ANSI_FIXED_FONT).GetFaceName() # "Lucida Console"
			self.editorSize = 12
		else:
			self.editorSize = int(self.editorSize)
		n = self.config.Read('DefaultNotation')
		if n:
			self.defaultNotation = n
			self.notations = [x for x in self.notations if x.id == n] + [x for x in self.notations if x.id != n]
		else:
			lang = i18n.getLang()
			self.defaultNotation = None
			if lang in defaultLangNotation:
				n = defaultLangNotation[lang].id
				self.notations = [x for x in self.notations if x.id == n] + [x for x in self.notations if x.id != n]
		self.config.SetPath('/AutoAdjust')
		spuriousLines = self.config.Read('spuriousLines')
		if spuriousLines:
			self.autoAdjustSpuriousLines = bool(int(spuriousLines))
		else:
			self.autoAdjustSpuriousLines = True
		tab2chordpro = self.config.Read('tab2chordpro')
		if tab2chordpro:
			self.autoAdjustTab2Chordpro = bool(int(tab2chordpro))
		else:
			self.autoAdjustTab2Chordpro = True
		easyKey = self.config.Read('easyKey')
		if easyKey:
			self.autoAdjustEasyKey = bool(int(easyKey))
		else:
			self.autoAdjustEasyKey = False
			self.notices['firstTimeEasyKey'] = True
		self.config.SetPath('/AutoAdjust/EasyChordsGroups')
		for k in easyChordsOrder:
			l = self.config.Read(k)
			if l:
				l = int(l)
			else:
				l = easyChords[k][2]
			self.SetEasyChordsGroup(k, l)
		self.config.SetPath('/App')
		ext = self.config.Read('defaultExtension')
		if not ext:
			self.defaultExtension = 'crd'
		else:
			self.defaultExtension = ext
		lang = self.config.Read('locale')
		if not lang:
			self.locale = None
		else:
			self.locale = lang
		self.config.SetPath('/AutoUpdate')
		f = self.config.Read('frequency')
		if not f:
			self.updateFrequency = 7
		else:
			self.updateFrequency = int(f)
		i = self.config.Read('ignored')
		if not i:
			self.ignoredUpdates = set()
		else:
			self.ignoredUpdates = set(i.split(','))
		u = self.config.Read('url')
		if not u:
			self.updateUrl = 'https://songpress.skeed.it/xmlrpc'
		else:
			self.updateUrl = u
		d = self.config.Read('lastCheck')
		if not d:
			self.updateLastCheck = None
		else:
			self.updateLastCheck = datetime.datetime.fromordinal(int(d))

	def Bool2String(self, param):
		return "1" if param else "0"

	def Save(self):
		self.config.SetPath('/Format')
		if self.chorusLabel is not None:
			self.config.Write('ChorusLabel', self.chorusLabel)
		self.config.Write('ShowChords', str(self.format.showChords))
		self.config.SetPath('/Format/Font')
		face = self.config.Write('Face', self.fontFace)
		self.config.SetPath('/Format/Style')
		self.config.Write('LabelVerses', self.Bool2String(self.labelVerses))
		self.config.SetPath('/Editor')
		self.config.Write('Face', self.editorFace)
		self.config.Write('Size', str(self.editorSize))
		if self.defaultNotation is not None:
			self.config.Write('DefaultNotation', self.defaultNotation)
		self.config.SetPath('/AutoAdjust')
		self.config.Write('spuriousLines', self.Bool2String(self.autoAdjustSpuriousLines))
		self.config.Write('tab2chordpro', self.Bool2String(self.autoAdjustTab2Chordpro))
		self.config.Write('easyKey', self.Bool2String(self.autoAdjustEasyKey))
		self.config.SetPath('/AutoAdjust/EasyChordsGroups')
		for k in easyChordsOrder:
			self.config.Write(k, str(self.GetEasyChordsGroup(k)))
		self.config.SetPath('/App')
		self.config.Write('defaultExtension', self.defaultExtension)
		if self.locale is not None:
			lang = self.config.Write('locale', self.locale)
		self.config.SetPath('/AutoUpdate')
		self.config.Write('frequency', str(self.updateFrequency))
		self.config.Write('ignored', ",".join(self.ignoredUpdates))
		self.config.Write('url', self.updateUrl)
		if self.updateLastCheck is not None:
			self.config.Write('lastCheck', str(self.updateLastCheck.toordinal()))

	def SetChorusLabel(self, c):
		self.chorusLabel = c
		self.decoratorFormat.SetChorusLabel(c)

	def SetDefaultNotation(self, notation):
		self.defaultNotation = notation
		self.notations = [x for x in self.notations if x.id == notation] + [x for x in self.notations if x.id != notation]

	def SetEasyChordsGroup(self, group, level):
		self.easyChordsGroup[group] = level
		self.easyChords = None

	def GetEasyChordsGroup(self, group):
		return self.easyChordsGroup[group]

	def GetEasyChords(self):
		if self.easyChords is None:
			self.easyChords = {'Eb': -1}
			for k in easyChords:
				chords = easyChords[k][1]
				l = self.easyChordsGroup[k] / 4.0
				for c in chords:
					if c in self.easyChords:
						self.easyChords[c] = max(self.easyChords[c], l)
					else:
						self.easyChords[c] = l

		return self.easyChords



###############################################################
# Name:			 PreviewCanvas.py
# Purpose:	 Window containing preview
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-02-21
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

import platform

import wx.adv

from .Renderer import *
from . import i18n

i18n.register('songpress.PreviewCanvas')


class PreviewCanvas(object):
	def __init__(self, parent, sf, notations, sd=SongDecorator(), embedded=False):
		object.__init__(self)
		show_link = not embedded and platform.system() != 'Linux'
		self.link = None
		if show_link:
			self.main_panel = wx.Window(parent)
			bSizer = wx.BoxSizer(wx.VERTICAL)
			self.link = wx.adv.HyperlinkCtrl(self.main_panel, 0, _("Copy formatted song to clipboard"), '')
			tt = wx.ToolTip(_("Copy formatted song to clipboard, so that it can be pasted in any program and printed"))
			self.link.SetToolTip(tt)
			bSizer.Add(self.link, 0, wx.EXPAND)
			parent = self.main_panel
		self.panel = wx.ScrolledWindow(parent, style=wx.BORDER_DOUBLE)
		self.panel.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
		self.pixedScrolled = 10
		self.panel.SetScrollbars(self.pixedScrolled, self.pixedScrolled, 0, 0)
		self.panel.Bind(wx.EVT_PAINT, self.OnPaint, self.panel)
		self.panel.SetBackgroundColour(wx.WHITE)
		self.text = ""
		if show_link:
			bSizer.Add(self.panel, 1, wx.EXPAND)
		else:
			self.main_panel = self.panel
		#SongFormat
		self.renderer = Renderer(sf, sd, notations)
		if show_link:
			self.main_panel.SetSizer(bSizer)
			self.main_panel.Layout()


	def OnPaint(self, e):
		#print("OnPaint")
		dc = wx.AutoBufferedPaintDC(self.panel)
		self.panel.PrepareDC(dc)
		dc.SetBackground(wx.WHITE_BRUSH)
		dc.Clear()
		w, h = self.renderer.Render(self.text, dc)
		self.panel.SetVirtualSize(wx.Size(int(w), int(h)))
		

	def Refresh(self, text):
		self.text = text
		self.panel.Refresh()

	def SetDecorator(self, sd):
		self.renderer.SetDecorator(sd)



###############################################################
# Name:			 SongDecorator.py
# Purpose:	 Base (and default) handlers for verse and chorus
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-02-21
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

from .SongFormat import *
from .SongBoxes import *


class SongDecorator(object):
	def __init__(self):
		object.__init__(self)
		self.dc = None
		# Current y
		self.y = 0
		self.dc = None
		# SongBox
		self.s = None
		
	def SetMarginText(self, text):
		# Modify text margins
		pass
		
	def SetMarginChord(self, chord):
		# Modify chord margins
		pass
		
	def SetMarginLine(self, line):
		# Modify line margins
		pass

	def SetMarginBlock(self, block):
		# Modify line margins
		pass

	def SetMarginSong(self, song):
		# Modify line margins
		pass
		
	def LayoutComposeLine(self, line):
		# Pass 1: determine size of text
		chordMaxH = 0
		chordMaxTH = 0
		textMaxH = 0
		textMaxTH = 0
		chordsOnly = True
		hasChords = False
		for t in line.boxes:
			self.dc.SetFont(t.font)
			text = t.text
			t.w, t.h = self.dc.GetTextExtent(text)
			if t.type == SongText.chord:
				hasChords = True
				self.SetMarginChord(t)
				chordMaxH = max(chordMaxH, t.h)
				chordMaxTH = max(chordMaxTH, t.GetTotalHeight())
			else:
				if text.strip() != '':
					chordsOnly = False
				self.SetMarginText(t)
				textMaxH = max(textMaxH, t.h)
				textMaxTH = max(textMaxTH, t.GetTotalHeight())
		if chordsOnly and hasChords:
			textMaxH = 0
			textMaxTH = 0
			line.textBaseline = chordMaxTH
		elif chordsOnly: # Block without text or chords
			w, h = self.dc.GetTextExtent(' ')
			textMaxH = h
			textMaxTH = h
			line.textBaseline = h
		else:
			line.textBaseline = chordMaxTH + chordMaxH * (line.parent.format.chordSpacing - 1) + textMaxTH
		line.chordBaseline = chordMaxTH
		
		line.h = line.textBaseline + textMaxH * (line.parent.format.textSpacing - 1)
		# Pass 2: set layout
		x = 0
		chordX = 0
		for t in line.boxes:
			self.dc.SetFont(t.font)
			if t.type == SongText.chord:
				t.x = max(x, chordX)
				x = t.x
				chordX = x + t.GetTotalWidth()
				t.y = line.chordBaseline - t.GetTotalHeight()
			else:
				t.x = x
				x_chord = max(x, chordX)
				x = t.x + t.GetTotalWidth()
				t.y = line.textBaseline - t.GetTotalHeight()
				if t.text.strip() == '':
					chordX = x_chord + t.GetTotalWidth()
			line.RelocateBox(t)
		self.SetMarginLine(line)
		
	def LayoutComposeBlock(self, block):
		y = 0
		for l in block.boxes:
			l.y = y
			y += l.GetTotalHeight()
			block.RelocateBox(l)
		self.SetMarginBlock(block)

	def LayoutComposeSong(self, song):
		y = 0
		self.dc.SetFont(song.format.wxFont)
		for b in song.boxes:
			b.y = y
			y += b.GetTotalHeight() + b.GetLastLineTextHeight() * song.format.blockSpacing
			song.RelocateBox(b)
		self.SetMarginSong(song)

	def LayoutCompose(self):
		# Postorder layout composing
		for block in self.s.boxes:
			for line in block.boxes:
				self.LayoutComposeLine(line)
		for block in self.s.boxes:
			self.LayoutComposeBlock(block)				
		self.LayoutComposeSong(self.s)
		
	def LayoutMoveBlock(self, block):
		# Move block within song
		pass
		
	def LayoutMoveLine(self, line):
		# Move line within block
		# If we need to, we can even move text and chords inside this line
		pass

	def LayoutMove(self):
		# Now that sizes are set, we can move elements inside each box if we need to
		for block in self.s.boxes:
			# Move block within song
			self.LayoutMoveBlock(block)
			for line in block.boxes:
				# Move line within block
				# If we need to, we can even move text and chords inside this line
				self.LayoutMoveLine(line)
				
	def PreDrawSong(self, song):
		pass
		
	def PreDrawBlock(self, block, bx, by):
		# bx, by: coordinates of top-left corner of drawable area
		pass
		
	def PreDrawLine(self, line, lx, ly):
		# lx, ly: coordinates of top-left corner of drawable area
		pass
		
	def PreDrawText(self, text, tx, ty):
		# tx, ty: coordinates of top-left corner of drawable area
		pass
		
	def DrawText(self, text, tx, ty):
		# tx, ty: coordinates of top-left corner of drawable area
		self.dc.SetFont(text.font)
		self.dc.SetTextForeground(text.color)
		self.dc.DrawText(text.text, int(tx + text.marginLeft), int(ty + text.marginTop))
		
	def PostDrawText(self, text, tx, ty):
		# tx, ty: coordinates of top-left corner of drawable area
		pass		
		
	def PostDrawLine(self, line, lx, ly):
		# lx, ly: coordinates of top-left corner of drawable area
		pass		
		
	def PostDrawBlock(self, block, bx, by):
		# bx, by: coordinates of top-left corner of drawable area
		pass
		
	def PostDrawSong(self, song):
		pass
		
	def DrawBoxes(self):
		if self.s.drawWholeSong:
			self.PreDrawSong(self.s)
			firstBlock = False
			firstBlockOffsetY = 0
		else:
			firstBlock = True
		for block in self.s.boxes:
			if block.drawBlock:
				if firstBlock:
					firstBlock = False
					firstBlockOffsetY = self.s.marginTop + block.y
				bx = self.s.marginLeft + block.x
				by = self.s.marginTop + block.y - firstBlockOffsetY
				self.PreDrawBlock(block, bx, by)
				for line in block.boxes:
					lx = bx + block.marginLeft + line.x
					ly = by + block.marginTop + line.y
					self.PreDrawLine(line, lx, ly)
					for text in line.boxes:
						tx = lx + line.marginLeft + text.x
						ty = ly + line.marginTop + text.y
						self.PreDrawText(text, tx, ty)
						self.DrawText(text, tx, ty)
						self.PostDrawText(text, tx, ty)
					self.PostDrawLine(line, lx, ly)
				self.PostDrawBlock(block, bx, by)
		if self.s.drawWholeSong:
			self.PostDrawSong(self.s)
		
	def InitDraw(self):
		pass
		
	def Draw(self, s, dc):
		# SongBox s
		self.s = s
		self.dc = dc
		self.InitDraw()
		self.LayoutCompose()
		self.LayoutMove()
		self.DrawBoxes()
		self.dc = None
		return self.s.GetTotalWidth(), self.s.GetTotalHeight()

	



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version Mar 10 2018)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
from . import i18n
i18n.register('songpress.NotationDialog')

###########################################################################
## Class NotationDialog
###########################################################################

class NotationDialog(wx.Dialog):

	def __init__(self, parent):
		wx.Dialog.__init__(self, parent, id=wx.ID_ANY, title=_(u"Change notation"), pos=wx.DefaultPosition,
											 size=wx.Size(361, 175), style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

		self.SetSizeHints(wx.DefaultSize, wx.DefaultSize)

		bSizer1 = wx.BoxSizer(wx.VERTICAL)

		bSizer2 = wx.BoxSizer(wx.HORIZONTAL)

		self.label1 = wx.StaticText(self, wx.ID_ANY, _(u"From notation:"), wx.DefaultPosition, wx.DefaultSize, 0)
		self.label1.Wrap(-1)
		bSizer2.Add(self.label1, 0, wx.ALL, 5)

		fromNotationChoices = []
		self.fromNotation = wx.Choice(self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, fromNotationChoices, 0)
		self.fromNotation.SetSelection(0)
		bSizer2.Add(self.fromNotation, 1, wx.ALL, 5)

		bSizer1.Add(bSizer2, 1, wx.EXPAND, 5)

		bSizer21 = wx.BoxSizer(wx.HORIZONTAL)

		self.label11 = wx.StaticText(self, wx.ID_ANY, _(u"To notation:"), wx.DefaultPosition, wx.DefaultSize, 0)
		self.label11.Wrap(-1)
		bSizer21.Add(self.label11, 0, wx.ALL, 5)

		toNotationChoices = []
		self.toNotation = wx.Choice(self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, toNotationChoices, 0)
		self.toNotation.SetSelection(0)
		bSizer21.Add(self.toNotation, 1, wx.ALL, 5)

		bSizer1.Add(bSizer21, 1, wx.EXPAND, 5)

		m_sdbSizer1 = wx.StdDialogButtonSizer()
		self.m_sdbSizer1OK = wx.Button(self, wx.ID_OK)
		m_sdbSizer1.AddButton(self.m_sdbSizer1OK)
		self.m_sdbSizer1Cancel = wx.Button(self, wx.ID_CANCEL)
		m_sdbSizer1.AddButton(self.m_sdbSizer1Cancel)
		m_sdbSizer1.Realize();

		bSizer1.Add(m_sdbSizer1, 0, wx.ALL | wx.EXPAND, 5)

		self.SetSizer(bSizer1)
		self.Layout()

		# Connect Events
		self.fromNotation.Bind(wx.EVT_CHOICE, self.OnFromNotation)

	def __del__(self):
		pass

	# Virtual event handlers, overide them in your derived class
	def OnFromNotation(self, event):
		event.Skip()





###############################################################
# Name:			 FontFaceDialog.py
# Purpose:	 wxWidgets dialog with a FontComboBox
#            selector
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-08-14
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

from .PreviewCanvas import *
from .FontComboBox import FontComboBox
from . import i18n

i18n.register('songpress.FontFaceDialog')

class FontFaceDialog(wx.Dialog):
	def __init__(self, parent, id, title, songFormat, songDecorator, decoratorFormat):
		self.format = songFormat
		self.originalFont = songFormat.face
		self.decoratorFormat = decoratorFormat
		wx.Dialog.__init__(self, parent, id, title, size=wx.Size(540, 320))

		bSizer1 = wx.BoxSizer(wx.VERTICAL)
		hSizer1 = wx.BoxSizer(wx.HORIZONTAL)
		
		m_staticText1 = wx.StaticText(self, wx.ID_ANY, _("Font"))
		m_staticText1.Wrap(-1);
		hSizer1.Add(m_staticText1, 0, wx.ALL, 5)
		
		self.fontCB = FontComboBox(self, wx.ID_ANY, songFormat.face)
		self.Bind(wx.EVT_COMBOBOX, self.OnFontSelected, self.fontCB)
		self.fontCB.Bind(wx.EVT_TEXT_ENTER, self.OnFontSelected, self.fontCB)
		self.fontCB.Bind(wx.EVT_KILL_FOCUS, self.OnFontSelected, self.fontCB)
		hSizer1.Add(self.fontCB, 1, wx.EXPAND | wx.ALL, 5)

		hSizer2 = wx.BoxSizer(wx.HORIZONTAL)
		
		m_staticText2 = wx.StaticText(self, wx.ID_ANY, _("Preview"))
		m_staticText2.Wrap(-1)
		hSizer2.Add(m_staticText2, 0, wx.ALL, 5)
		
		self.previewCanvas = PreviewCanvas(self, songFormat, songDecorator, embedded=True)
		hSizer2.Add(self.previewCanvas.panel, 1, wx.EXPAND | wx.ALL, 5)
		previewSong = _("{t:My Bonnie}\n\nMy [D]Bonnie lies [G]over the [D]ocean\noh [G]bring back my [A]Bonnie to [D]me!\n\n{soc}\n[D]Bring back, [E-]bring back,\n[A]bring back my Bonnie to [D]me!\n{eoc}")
		self.previewCanvas.Refresh(previewSong)
		
		bSizer1.Add(hSizer1, 0, wx.EXPAND, 5)
		bSizer1.Add(hSizer2, 1, wx.EXPAND, 5)
		
		bSizer2 = wx.BoxSizer(wx.HORIZONTAL)
		
		bSizer2.AddStretchSpacer(1)
		
		self.ok = wx.Button(self, wx.ID_OK, _("OK"))
		bSizer2.Add(self.ok, 0, wx.ALL, 5 )
		
		self.cancel = wx.Button(self, wx.ID_CANCEL, _("Cancel"))
		bSizer2.Add(self.cancel, 0, wx.ALL, 5)
		self.Bind(wx.EVT_BUTTON, self.OnCancel, self.cancel)
		
		bSizer1.Add(bSizer2, 0, wx.EXPAND, 5)
		
		self.SetSizer(bSizer1)
		self.Layout()
		
	def SetFont(self, font):
		self.format.face = font
		self.format.chord.face = font
		self.format.comment.face = font
		self.format.chorus.face = font
		self.format.chorus.chord.face = font
		self.format.chorus.comment.face = font
		for v in self.format.verse:
			v.face = font
			v.chord.face = font
			v.comment.face = font
		self.format.title.face = font
		self.decoratorFormat.face = font
		self.decoratorFormat.chorus.face = font
		previewSong = _("{t:My Bonnie}\n\nMy [D]Bonnie lies [G]over the [D]ocean\noh [G]bring back my [A]Bonnie to [D]me!\n\n{soc}\n[D]Bring back, [E-]bring back,\n[A]bring back my Bonnie to [D]me!\n{eoc}")		
		self.previewCanvas.Refresh(previewSong)	
		
	def OnFontSelected(self, evt):
		font = self.fontCB.GetValue()
		self.SetFont(font)
		evt.Skip()
		
	def GetValue(self):
		return self.fontCB.GetValue()
		
	def OnCancel(self, evt):
		font = self.originalFont
		self.SetFont(font)
		evt.Skip()
		



###############################################################
# Name:			 songimpress.py
# Purpose:	 Generate PowerPoint presentation from songs
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2019-02-02
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

import sys

from pptx import Presentation


class SongPresentation:
	def __init__(self, template, out_filename):
		self.pres = Presentation(template)
		self.layout = self.pres.slide_layouts[3]
		self.prev = None
		self.out_filename = out_filename

	def _add_slide(self, cur, next):
		slide = self.pres.slides.add_slide(self.layout)
		ph = list(slide.placeholders)
		t1 = ph[1]
		t2 = ph[2]
		t1.text = cur
		t2.text = next

	def _add_empty_slide(self):
		self.pres.slides.add_slide(self.layout)

	def add_line(self, line):
		if self.prev is not None:
			self._add_slide(self.prev, line)
		self.prev = line

	def end_song(self):
		if self.prev is not None:
			self._add_slide(self.prev, '')
		self._add_empty_slide()
		self.prev = None

	def close(self):
		if self.prev is not None:
			self._add_slide(self.prev, '')
		self.pres.save(self.out_filename)

	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.close()


def to_presentation(lines, output_file, template_file):
	with SongPresentation(template_file, output_file) as c:
		for line in lines:
			line = line.replace('\n', '').strip()
			if line == '---':
				c.end_song()
			elif line != '':
				c.add_line(line)


if __name__ == '__main__':
	if len(sys.argv) != 4:
		print("Usage: songimpress.py SONG_FILE OUTPUT_PRESENTATION TEMPLATE_PRESENTATION")
		exit(1)

	with open(sys.argv[1]) as f:
		to_presentation(f, sys.argv[2], sys.argv[3])



###############################################################
# Name:			 MyTransposeDialog.py
# Purpose:	 Transposing dialog
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-12-02
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

from .TransposeDialog import *
from .Transpose import *


class MyTransposeDialog(TransposeDialog):
	def __init__(self, parent, notations, text):
		TransposeDialog.__init__(self, parent)
		self.notations = notations
		self.text = text
		fn = autodetectNotation(text, notations)
		self.fk = autodetectKey(text, fn)
		for n in self.notations:
			i = self.notation.Append(n.desc)
			self.notation.SetClientData(i, n)
			if n == fn:
				self.notation.SetSelection(i)
		self.OnNotation()
			
	def OnNotation(self, event=None):
		i = self.notation.GetSelection()
		n = self.notation.GetClientData(i)
		self.fromKey.Clear()
		self.toKey.Clear()
		for k in orderedKeys:
			kn = "%s / %s" % (
				translateChord(k, dNotation=n),
				translateChord(scales[k][1][5] + "m", dNotation=n),
			)
			i = self.fromKey.Append(kn)
			self.fromKey.SetClientData(i, k)
			if chord2pos(self.fk) == chord2pos(k):
				self.fromKey.SetSelection(i)
			i = self.toKey.Append(kn)
			self.toKey.SetClientData(i, k)
		self.UpdateToKey()
		if event is not None:
			event.Skip()
			
	def UpdateToKey(self):
		try:
			s = int(self.semitones.GetValue())
		except:
			return
		self.toKey.SetSelection((self.fromKey.GetSelection() + s) % 12)
		
	def GetTransposed(self):
		return transposeChordPro(
			self.fromKey.GetClientData(self.fromKey.GetSelection()),
			self.toKey.GetClientData(self.toKey.GetSelection()),
			self.text,
			self.notation.GetClientData(self.notation.GetSelection())
		)
		
	def OnSemitones(self, evt):
		self.UpdateToKey()
		evt.Skip()
		
	def OnFromKey(self, evt):
		self.UpdateToKey()
		evt.Skip()
		
	def OnToKey(self, evt):
		g = (self.toKey.GetSelection() - self.fromKey.GetSelection()) % 12
		if g > 7:
			g = -12 + g
		self.semitones.SetValue(g)
		evt.Skip()



###############################################################
# Name:			 i18n.py
# Purpose:	 Support internationalization
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-09-11
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

import inspect
import os
import locale

from .Globals import *


registered_modules = []
# 2-character language code of current language
current_language = None
# wxPython wx.Locale object for current language
mylocale = None
# 2-character language code of default language, set during init
defaultLang = None
# Array mapping 2-character language codes of languages supported by Songpress
# to language names. To be set during init phase, by calling init function
supportedLangs = None


"""
Client shall:
1. Call `init`
2. Call either `setLang` (if a langugage is chosen by the user)
   or `setSystemLang`
"""


def init(default_lang, supported_langs):
	global defaultLang
	global supportedLangs
	defaultLang = default_lang
	supportedLangs = supported_langs


_ = lambda x: x


def setLang(l):
	global current_language, mylocale
	current_language = l
	langid = wx.Locale.FindLanguageInfo(l).Language
	mylocale = wx.Locale(langid)
	localedir = os.path.join(glb.path, "locale")
	mylocale.AddCatalogLookupPathPrefix(localedir)
	for mod in registered_modules:
		domain = mod.__name__
		mylocale.AddCatalog(domain)


def getLang():
	return current_language


def setSystemLang():
	l = locale.getdefaultlocale()
	if l is not None and l[0][:2] in supportedLangs:
		setLang(l[0][:2])
	else:
		setLang(defaultLang)


def register(moduleName=None):
	if moduleName is None:
		# Try to determine caller module name
		frm = inspect.stack()[1]
		mod = inspect.getmodule(frm[0])
		moduleName = mod.__name__
	else:
		mod = sys.modules[moduleName]
	registered_modules.append(mod)
	mod._ = wx.GetTranslation


def localizeXrc(filename):
	if current_language != defaultLang:
		d, domain = os.path.split(filename)
		localedir = os.path.join(glb.AddPath(d), "locale")
		mylocale.AddCatalogLookupPathPrefix(localedir)
		mylocale.AddCatalog(domain)



###############################################################
# Name:			 SoongBoxes.py
# Purpose:	 Elements that make up a song
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-02-21
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

import wx

from .SongFormat import *


class SongBox(object):
	def __init__(self, x, y, w, h):
		object.__init__(self)
		self.x = x
		self.y = y
		self.w = w
		self.h = h
		self.marginLeft = 0
		self.marginRight = 0
		self.marginTop = 0
		self.marginBottom = 0
		self.boxes = []

	def RelocateBox(self, box):
		self.w = max(self.w, box.x + box.GetTotalWidth())
		self.h = max(self.h, box.y + box.GetTotalHeight())
		
	def AddBox(self, box):
		self.boxes.append(box)
		box.parent = self
		self.RelocateBox(box)
		
	def SetMargin(self, top, right, bottom, left):
		self.marginTop = top
		self.marginRight = right
		self.marginBottom = bottom
		self.marginLeft = left
		
	def GetTotalHeight(self):
		return self.h + self.marginTop + self.marginBottom

	def GetTotalWidth(self):
		return self.w + self.marginLeft + self.marginRight

	
class SongSong(SongBox):
	def __init__(self, format):
		SongBox.__init__(self, 0, 0, 0, 0)
		self.format = format
		self.verseCount = 0
		self.chorusCount = 0
		self.labelCount = 0
		self.drawWholeSong = False


class SongBlock(SongBox):
	# types
	verse = 1
	chorus = 2
	title = 3
	subtitle = 4

	def __init__(self, type, format):
		SongBox.__init__(self, 0, 0, 0, 0)
		self.type = type
		self.verseNumber = 0
		self.format = format
		self.drawBlock = False
		self.label = None
		self.chords = []
		
	def RemoveChordBoxes(self):
		for l in self.boxes:
			l.RemoveChordBoxes()

	def GetLastLineTextHeight(self):
		if len(self.boxes) > 0:
			return self.boxes[-1].GetTextHeight()
		return 0

			
class SongLine(SongBox):
	def __init__(self):
		SongBox.__init__(self, 0, 0, 0, 0)
		self.hasChords = False
		self.chordBaseline = 0
		self.textBaseline = 0

	def AddBox(self, text):
		if text.type == text.chord:
			self.hasChords = True
		SongBox.AddBox(self, text)
		
	def RemoveChordBoxes(self):
		self.boxes = [b for b in self.boxes if b.type != b.chord]

	def GetTextHeight(self):
		tbs = [b for b in self.boxes if b.type != b.chord]
		if len(tbs) == 0:
			return self.h
		return max(tb.h for tb in tbs)

	
class SongText(SongBox):
	text = 1
	chord = 2
	comment = 3
	title = 4
	subtitle = 5
	
	def __init__(self, text, font, type, color):
		SongBox.__init__(self, 0, 0, 0, 0)
		self.text = text
		self.font = font
		self.type = type
		self.color = color



###############################################################
# Name:			 MyListDialog.py
# Purpose:	 Generic dialog box asking for an item in a list
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2019-02-02
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

import wx

from .ListDialog import ListDialog


class MyListDialog(ListDialog):
	def __init__(self, parent, message, title, elements):
		"""
		Show a dialog box asking to select an item from a list

		:param message: message to be displayed
		:param title: dialog title
		:param elements: list of strings: items to be chosen from
		"""
		super().__init__(parent)
		self.SetTitle(title)
		self.label.SetLabel(message)
		self.elements_list = elements
		self.elements.InsertItems(elements, 0)

	def GetSelectedIndex(self):
		return self.elements.GetSelection()

	def GetSelectedString(self):
		return self.elements_list[self.elements.GetSelection()]




###############################################################
# Name:			 StandardVerseNumbers.py
# Purpose:	 Decorator that adds verse numbering
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-03-14
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

from songpress.SongDecorator import *
from songpress.SongFormat import *
import wx

class Format(FontFormat):
	def __init__(self, sf, chorusLabel):
		FontFormat.__init__(self)
		self.face = sf.face
		self.size = sf.size
		self.bold = sf.bold
		self.italic = sf.italic
		self.underline = sf.underline
		self.leftMargin = 0.5
		self.rightMargin = 0.5
		self.leftPadding = 0.25
		self.rightPadding = 0.25
		self.topPadding = 0.1
		self.bottomPadding = 0.1
		self.chorus = FontFormat()
		self.chorus.face = sf.chorus.face
		self.chorus.size = sf.chorus.size
		self.chorus.bold = sf.chorus.bold
		self.chorus.italic = sf.chorus.italic
		self.chorus.underline = sf.chorus.underline
		self.chorus.label = chorusLabel

	def SetChorusLabel(self, label):
		self.chorus.label = label

	def GetChorusLabel(self):
		return self.chorus.label


class Decorator(SongDecorator):
	def __init__(self, format):
		SongDecorator.__init__(self)
		self.format = format

	wxBlack = wx.Colour(0, 0, 0)
	wxWhite = wx.Colour(255, 255, 255)
	wxGrey = wx.Colour(200, 200, 200)

	def SetMarginBlock(self, block):
		font = block.format.wxFont
		self.dc.SetFont(font)
		baseWidth, baseHeight = self.dc.GetTextExtent("0")
		if block.type == block.verse:
			text = block.label if block.label is not None else ''
			text_spacer = text if text != '' else str(self.s.labelCount)
			w, h = self.dc.GetTextExtent(text_spacer)

			w += baseWidth * (
				self.format.leftMargin + self.format.leftPadding + self.format.rightMargin + self.format.rightPadding
			)
		elif block.type == block.title:
			w = 0
		else:
			if block.label is not None:
				text = block.label
			else:
				text = self.format.chorus.label
			w, h = self.dc.GetTextExtent(text)
			w += baseWidth * (
				self.format.leftMargin + self.format.leftPadding + self.format.rightMargin + self.format.rightPadding
			)
		block.SetMargin(0, 0, 0, w)

	def PreDrawBlock(self, block, bx, by):
		if block.type != block.title and len(block.boxes) > 0:
			font = block.format.wxFont
			self.dc.SetFont(font)
			baseWidth, baseHeight = self.dc.GetTextExtent("0")
			if block.type == block.verse:
				background = self.wxGrey
				foreground = self.wxBlack
				if block.label is not None:
					text = block.label
					text_spacer = text
					if text == '':
						text_spacer = str(self.s.labelCount)
				else:
					text = str(block.verseLabelNumber)
					text_spacer = str(self.s.labelCount)
				w, h = self.dc.GetTextExtent(text_spacer)

			else:
				background = self.wxBlack
				foreground = self.wxWhite
				if block.label is not None:
					text = block.label
				else:
					text = self.format.chorus.label
				w, h = self.dc.GetTextExtent(text)
			if text != '':
				realW, realH = self.dc.GetTextExtent(text)
				rx = bx + self.format.leftMargin * baseWidth
				tx = (rx
					+ baseWidth * self.format.leftPadding
					+ 0.5 * (w - realW))
				ty = by + block.boxes[0].textBaseline + block.marginTop - h
				ry = ty - self.format.topPadding * baseHeight
				brush = wx.Brush(background, wx.SOLID)
				self.dc.SetBrush(brush)
				self.dc.DrawRectangle(int(rx), int(ry),
					int(w + baseWidth * (self.format.leftPadding + self.format.rightPadding)),
					int(h + baseHeight * (self.format.topPadding + self.format.bottomPadding)))
				brush = wx.Brush(foreground, wx.SOLID)
				self.dc.SetBrush(brush)
				self.dc.SetTextForeground(foreground)
				self.dc.SetBackgroundMode(wx.TRANSPARENT)
				self.dc.DrawText(text, int(tx), int(ty))
				self.dc.SetTextForeground(self.wxBlack)



###############################################################
# Name:			 SongTokenizer.py
# Purpose:	 Subclass of Tokenizer to tokenize songs
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-01-31
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

		
import re

from .Tokenizer import *


class SongTokenizer(Tokenizer):
	def __init__(self, l):
		Tokenizer.__init__(self, l)
		
	#Tokens
	openCurlyToken = TokenType('(\{)', 'openCurlyToken')
	closeCurlyToken = TokenType('(\})', 'closeCurlyToken')
	commentToken = TokenType('^([ \t]*#.*)$', 'commentToken')
	normalToken = TokenType('([^[{]+)', 'normalToken')
	commandToken = TokenType('([^}:]+)', 'commandToken')
	attrToken = TokenType('([^}]+)', 'attrToken')
	chordToken = TokenType('(\[[^]]*)', 'chordToken')
	closeChordToken = TokenType('(\])', 'closeChordToken')
	colonToken = TokenType('(:)', 'colonToken')

	#States
	normal = 1
	command = 2
	attr = 3
	
	#Transition function
	transition = {
		normal: (
			(openCurlyToken, command),
			(chordToken, normal),
			(closeChordToken, normal),
			(commentToken, normal),
			(normalToken, normal),
		),
		command: (
			(colonToken, attr),
			(closeCurlyToken, normal),
			(commandToken, command),
		),
		attr: (
			(attrToken, attr),
			(closeCurlyToken, normal),
			(commentToken, normal),
		)
	}




###############################################################
# Name:			 HTML.py
# Purpose:	 Export song to HTML
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2010-05-05
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

from .Renderer import *

template = """
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<head>
		<title>%(title)s</title>
		<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
		<style type="text/css">
			<!--
				div.chorus,div.verse,div.title,div.titleblock {
					margin-bottom: 1.5em;
					font-family: "%(face)s";
				}
				div.chorus table,div.verse table, div.title table {
					margin: 0px;
					padding: 0px;
					border: 0px;
					border-collapse: collapse;
				}
				.chorus {
					font-weight: bold;
				}
				.chord {
					font-style: italic;
					font-size: small;
					font-weight: normal;
				}
				.title {
					font-weight: bold;
					text-decoration: underline;
				}
				.subtitle {
					font-weight: normal;
					font-style: italic;
					font-size: 0.95em;
				}
			-->
		</style>
	</head>
	<body>
		%(body)s
	</body>
</html>
"""

class HtmlExporter(object):
	def __init__(self, sf):
		object.__init__(self)
		self.sf = sf
		self.out = ""

	def Draw(self, song, dc):
		# dc is a dummy parameter
		classes = {
			SongBlock.title: 'titleblock',
			SongBlock.chorus: 'chorus',
			SongBlock.verse: 'verse',
		}
		out = ""
		title = "Songpress"
		for block in song.boxes:
			b = "<div class=\"%s\">" % (classes[block.type])
			for line in block.boxes:
				tc = "<td>"
				tt = "<td>"
				new_line = True
				for t in line.boxes:
					if t.type == SongText.title:
						tt += f"""<span class="title">{t.text.replace(" ", "&nbsp;")}</span>"""
					elif t.type == SongText.subtitle:
						tt += f"""<span class="subtitle">{t.text.replace(" ", "&nbsp;")}</span>"""
					elif t.type == SongText.chord:
						if not new_line:
							tc += "</td><td>"
							tt += "</td><td>"
						tc += t.text.replace(" ", "&nbsp;")
					else:
						tt += t.text.replace(" ", "&nbsp;")
						if block.type == SongBlock.title:
							title = t.text
					new_line = False
				tc += "</td>"
				tt += "</td>"
				b += "<table cellpadding=\"0\">\n<tr class=\"chord\">%s</tr>\n<tr>%s</tr>\n</table>\n" % (tc, tt)
			b += "</div>"
			out += b
		self.out = template % {
				'title': title,
				'body': out,
				'face': self.sf.face,
			}
		return 0, 0

	def getHtml(self):
		return self.out


class TabExporter(object):
	def __init__(self, sf):
		object.__init__(self)
		self.sf = sf
		self.out = ""

	def Draw(self, song, dc):
		# dc is a dummy parameter
		out = []
		for block in song.boxes:
			for line in block.boxes:
				tc = ""
				tt = ""
				spaces = 0
				for t in line.boxes:
					if t.type == SongText.chord:
						# n = spaces + t.text
						if spaces >= 0:
							tc += " " * spaces
						else:
							tt += " " * -spaces
						tc += t.text
						spaces = -len(t.text)
					else:
						tt += t.text
						spaces += len(t.text)
				if tc != "":
					out.append(tc)
				out.append(tt)
			out.append("\n")
		self.out = "\n".join(out).strip()
		return 0, 0

	def getTab(self):
		return self.out



###############################################################
# Name:			 MyNotationDialog.py
# Purpose:	 Dialog to change notation
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-12-09
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

from .NotationDialog import *
from .Transpose import *


class MyNotationDialog(NotationDialog):
	def __init__(self, parent, notations, text):
		NotationDialog.__init__(self, parent)
		self.notations = notations
		fn = autodetectNotation(text, notations)
		for n in self.notations:
			i = self.fromNotation.Append(n.desc)
			self.fromNotation.SetClientData(i, n)
			if n == fn:
				self.fromNotation.SetSelection(i)
			i = self.toNotation.Append(n.desc)
			self.toNotation.SetClientData(i, n)
		self.toNotation.SetSelection(0)
		self.OnFromNotation(None)
		self.text = text
			
	def OnFromNotation(self, event):
		i = self.fromNotation.GetSelection()
		if self.toNotation.GetSelection() == i:
			self.toNotation.SetSelection(1 if i == 0 else 0) 
		
	def ChangeChordNotation(self):
		return translateChordPro(
			self.text,
			self.fromNotation.GetClientData(self.fromNotation.GetSelection()),
			self.toNotation.GetClientData(self.toNotation.GetSelection())
		)



###############################################################
# Name:			 Pref.py
# Purpose:	 Preference management
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-05-01
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

import copy
import string
from xml.dom import minidom


class Property(object):
	"""Structure for properties"""
	
	def __getValue(self):
		return self.__value
		
	def __setValue(self, value):
		if self.validator != None:
			valid, value = self.validator(value)
			if valid:
				self.__value = value
		else:
			self.__value = value
		if self.gui != None:
			self.gui.Changed()

	value = property(__getValue, __setValue)
	
	def __init__(self, value, validator = None, gui = None):
		object.__init__(self)
		self.validator = validator
		self.gui = gui(self) if gui != None else None
		self.value = value
	
		
class PropertyDef(object):
	"""Structure for property definitions"""
	def __init__(self, name, ptype, init = None, validator = None, gui = None):
		object.__init__(self)
		self.name = name
		self.ptype = ptype
		self.init = init
		self.validator = validator
		self.gui = gui


class XmlSerializer(object):
	def __init__(self):
		object.__init__(self)
		self.idn = 0
		self.dom = minidom.Document()
		self.ids = {}
		
	def RegisterId(self, p):
		self.idn += 1
		self.ids[p] = self.idn
		return str(self.idn)
		
	def GetId(self, p):
		return str(self.ids[p])
		
	def Serialize(self, p):
		self.dom.appendChild(p.XmlSerialize(self))
				
class XmlDeserializer(object):
	def __init__(self, dom, classes):
		object.__init__(self)
		self.dom = dom
		self.ids = {}
		self.classes = dict((c.__name__, c) for c in classes)
		
	def RegisterId(self, id, p):
		self.ids[id] = p
		
	def GetById(self, p):
		return self.ids[p]
		
	def Serialize(self, p):
		self.dom.appendChild(p.XmlSerialize(self))
		
	def GetClass(self, name):
		return self.classes[name]
		
	def Deserialize(self):
		l = []
		for el in self.dom.childNodes:
			l.append(Preferences.XmlDeserialize(self, el))
		return l
		
		
class NameNotFoundException(Exception):
	def __init__(self, arg):
		Exception.__init__(self, arg)


class Preferences(object):
	"""Class for structured property objects."""
	def __init__(self, parents = [], gui = None):
		self.__prefs = {}
		object.__init__(self)
		self.__parents = dict((x, x) for x in parents)
		self.__parentsByName = dict((x, x) for x in parents)
		self.__owner = None
		self.__CacheParents()
		self.__gui = gui
		for n, v in self.__GetAllPrefDef():
			f = v.init
			if f != None:
				self.__setattr__(n, f())
		
	prefDef = {}
	
	@classmethod
	def Register(cls, name, ptype, init = None, validator = None, gui = None):
		if not ('prefDef' in cls.__dict__):
			cls.prefDef = {}
		cls.prefDef[name] = PropertyDef(name, ptype, init, validator, gui)
		
	def __SetOwner(self, owner):
		self.__owner = owner
		self.__CacheParents()
		
	def __CacheParents(self):
		def Navigate(node, path):
			if node == None:
				return None
			if path == '..':
				return node.__owner
			if path in node.__prefs:
				return node.__prefs.value
			return None
		for n in self.__parentsByName:
			if type(n) == str:
				del self.__parents[self.__parentsByName[n]]
				val = reduce(Navigate, string.split(n, '/'), self)
				self.__parentsByName[n] = val
				self.__parents[val] = n		
		
	def __GetPrefByName(self, name):
		if name in self.__prefs:
			return self.__prefs[name]
		for p in self.__parents:
			try:
				return p.__GetPrefByName(name)
			except NameNotFoundException:
				pass
		raise NameNotFoundException(name)
	
	@classmethod
	def __GetPrefDefByName(cls, name):
		if name in cls.prefDef:
			return cls.prefDef[name]
		for b in cls.__bases__:
			try:
				return b.__GetPrefDefByName(name)
			except Exception:
				pass
		raise NameNotFoundException(name)
	
	def __getattr__(self, name):
		return self.__GetPrefByName(name).value
		
	def __setattr__(self, name, value):
		if name == '_Preferences__prefs':
			object.__setattr__(self, name, value)
			return
		try:
			pd = self.__GetPrefDefByName(name)
			assert isinstance(value, pd.ptype)
			if not name in self.__prefs:
				self.__prefs[name] = Property(value, pd.validator, pd.gui)
			else:
				self.__prefs[name].value = value
			if isinstance(value, Preferences):
				value.__SetOwner(self)
		except NameNotFoundException:
			object.__setattr__(self, name, value)
	
	def GetInherit(self, name):
		return not (name in self.__prefs)
		
	def SetInherit(self, name, inherit):
		if inherit:
			if name in self.__prefs:
				del self.__prefs[name]
		else:
			p = self.__GetPrefByName(name)
			self.__prefs[name] = Preferences.__DeepCopyPrefProp(p, None, self)
			
	def __GetAllPrefs(self):
		for n in self.__prefs:
			yield n, self.__prefs[n]
		for p in self.__parents:
			for n, v in p.__GetAllPrefs():
				yield n, v
	
	@classmethod
	def __GetAllPrefDef(cls):
		if 'prefDef' in cls.__dict__:	
			for n in cls.prefDef:
				yield n, cls.prefDef[n]
		for b in cls.__bases__:
			if isinstance(b, Preferences):
				for n, v in b.__GetAllPrefDef():
					yield n, v
	
	@staticmethod
	def __DeepCopyPrefProp(p, d, owner):
		q = Property(
			Preferences.__DeepCopyPref(p.value, d, owner) if isinstance(p.value, Preferences) else copy.deepcopy(p.value, d),
			copy.deepcopy(p.validator, d),
			copy.deepcopy(p.gui, d)
		)
		return q
	
	@staticmethod 
	def __DeepCopyPref(p, d, owner):
		c = Preferences(
			[pa for pa in p.__parentsByName],
			copy.deepcopy(p.__gui, d)
		)
		c.__owner = owner
		c.__CacheParents()
		for n in p.__prefs:
			c.__prefs[n] = Preferences.__DeepCopyPrefProp(p.__prefs[n], p)
		return c
		
	def XmlSerialize(self, xm):
		xd = xm.dom
		pref = xd.createElement("pref")
		pref.setAttribute("ptype", self.__class__.__name__)
		pref.setAttribute("id", xm.RegisterId(self))
		for n in self.__parentsByName:
			par = xd.createElement("parent")
			if type(n) == str:
				par.setAttribute("rel", n)
			else:
				par.setAttribute("id", xm.GetId(n))
			pref.appendChild(par)
		for n in self.__prefs:
			elem = xd.createElement("elem")
			elem.setAttribute("name", n)
			p = self.__prefs[n].value
			if isinstance(p, Preferences):
				elem.appendChild(p.XmlSerialize(xm))
			else:
				t = xd.createElement(type(p).__name__)
				t.appendChild(xd.createTextNode(str(p)))
				elem.appendChild(t)
			pref.appendChild(elem)
		return pref
		
	@staticmethod
	def XmlDeserialize(xm, pref):
		while pref.nodeName == '#text':
			pref = pref.nextSibling
		if pref.nodeName == 'pref':
			p = [c.getAttribute('rel') if c.hasAttribute('rel') else xm.GetById(c.getAttribute('id')) for c in pref.childNodes if c.nodeName == 'parent']
			c = xm.GetClass(pref.getAttribute('ptype'))
			assert issubclass(c, Preferences)
			pr = c(p)
			xm.RegisterId(pref.getAttribute('id'), pr)
			for c in pref.childNodes:
				if c.nodeName == 'elem':
					name = c.getAttribute('name')
					pr.__setattr__(name, pr.XmlDeserialize(xm, c.firstChild))
			return pr
		elif pref.nodeName == 'str':
			return str(pref.firstChild.nodeValue)
		elif pref.nodeName == 'int':
			return int(pref.firstChild.nodeValue)
			
			




# -*- coding: iso-8859-1 -*-

###############################################################
# Name:			 Transpose.py
# Purpose:	 Transposing services
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-11-18
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

from . import i18n
import re
import math

i18n.register('songpress.Transpose')


class Notation(object):
	def __init__(self, id, desc, chords, repl, replrev):
		object.__init__(self)
		self.id = id
		self.descv = desc
		self.chords = chords
		self.chordDict = {}
		i = 0
		for k in chords:
			self.chordDict[k.upper()] = i
			i += 1
		self.repl = [(re.compile(x[0]), x[1]) for x in repl]
		self.replrev = [(re.compile(x[0]), x[1]) for x in replrev]

	def GetDesc(self):
		return _(self.descv)

	def SetDesc(self, v):
		pass

	desc = property(GetDesc, SetDesc)

	def Ord2Chord(self, pos):
		return self.chords[pos]

	def Chord2Ord(self, chord):
		return self.chordDict[chord.upper()]

	def __AlterationStandard(self, a, rs):
		for r in rs:
			p = 0
			b = ''
			for m in r[0].finditer(a):
				b += a[p:m.start()] + r[1]
				p = m.end()
			b += a[p:]
			a = b
		return a

	def PostprocessingFromStandard(self, c, a):
		return c, a

	def PreprocessingToStandard(self, chord):
		return chord

	def AlterationFromStandard(self, a):
		return self.__AlterationStandard(a, self.repl)

	def AlterationToStandard(self, a):
		return self.__AlterationStandard(a, self.replrev)


enNotation = Notation(
	"enNotation",
	_("American (C D E... B)"),
	['C', 'D', 'E', 'F', 'G', 'A', 'B'],
	[],
	[]
)

itNotation = Notation(
	"itNotation",
	_("Italian (Do Re Mi... Si)"),
	['Do', 'Re', 'Mi', 'Fa', 'Sol', 'La', 'Si'],
	[
		(r'maj7', '7+'),
		(r'sus4', '4'),
		(r'^m', '-')
	],
	[
		(r'7\+', 'maj7'),
		(r'^4', 'sus4'),
		(r'^-', 'm')
	]
)

frNotation = Notation(
	"frNotation",
	_(u"French (Do Ré Mi... Si)"),
	['Do', u'Ré', 'Mi', 'Fa', 'Sol', 'La', 'Si'],
	[
		(r'maj7', '7+'),
		(r'sus4', '4'),
	],
	[
		(r'7\+', 'maj7'),
		(r'^4', 'sus4'),
	]
)

ptNotation = Notation(
	"ptNotation",
	_(u"Portuguese (Dó Ré Mi... Si)"),
	[u'Dó', u'Ré', 'Mi', u'Fá', 'Sol', u'Lá', 'Si'],
	[
		(r'maj7', '7+'),
		(r'sus4', '4'),
		(r'^m', '-')
	],
	[
		(r'7\+', 'maj7'),
		(r'^4', 'sus4'),
		(r'^-', 'm')
	]
)


easyChords = {
	'basic': (_("Basic chords (A, E, D)"), ["A", "E", "D"], 4),
	'Cprog': (_("50s progr. in C (C, Am, Dm, G7)"), ["C", "Am", "Dm", "G7"], 4),
	'F': (_("F chord"), ["F"], 2),
	'Gprog': (_("50s progr. in G (G, Em, Am, D7)"), ["G", "Em", "Am", "D7"], 4),
	'Dprog': (_("50s progr. in D (D, Bm, Em, A7)"), ["D", "Bm", "Em", "A7"], 1),
	'Aprog': (_("50s progr. in A (A, F#m, Bm, E7)"), ["A", "F#m", "Bm", "E7"], 1),
	'C#(m)(7)': (_("C#, C#m, C#7 chords"), ["C#", "C#m", "C#7"], 0),
	'Fprog': (_("50s progr. in F (F, Dm, Gm, C7)"), ["F", "Dm", "Gm", "C7"], 0),
	'BB7': (_("B and B7 chords"), ["B", "B7"], 0),
}

def getEasyChordsDescription(e):
	return _(e[0])

easyChordsOrder = ['basic', 'Cprog', 'F', 'Gprog', 'Dprog', 'Aprog', 'C#(m)(7)', 'Fprog', 'BB7']


class GermanNotation(Notation):
	def PreprocessingToStandard(self, chord):
		if chord.lower().startswith('bb'):
			return "Hb" + chord[2:]
		return chord

	def PostprocessingFromStandard(self, c, a):
		if c == 'Hb':
			c = 'Bb'
		return c, a


trad_de_map = [
	('C#', 'Cis'),
	('Db', 'Des'),
	('Eb', 'Es'),
	('F#', 'Fis'),
	('Gb', 'Ges'),
	('G#', 'Gis'),
	('Ab', 'As'),
	('A#', 'Ais'),
	('Hb', 'B'),
]

trad_de_from = {a.lower(): b for a, b in trad_de_map}
trad_de_to = [(b.lower(), a) for a, b in trad_de_map]


class TraditionalGermanNotation(Notation):
	def PreprocessingToStandard(self, chord):
		cl = chord.lower()
		for a, b in trad_de_to:
			if cl.startswith(a):
				return b + chord[len(a):]
		return chord

	def PostprocessingFromStandard(self, c, a):
		c = trad_de_from.get(c.lower(), c)
		if a != "" and a[0] == 'm' and not a.startswith('maj'):
			c = c.lower()
		return c, a


deNotation = GermanNotation(
	"deNotation",
	_("German/Sacandinavian (C C#/Db D... Bb H)"),
	['C', 'D', 'E', 'F', 'G', 'A', 'H'],
	[],
	[]
)


tradDeNotation = TraditionalGermanNotation(
	"tradDeNotation",
	_("Traditional German (C Cis/Des D... B H)"),
	['C', 'D', 'E', 'F', 'G', 'A', 'H'],
	[
		('C#', 'Cis'),
		('Db', 'Des'),
	],
	[
		('C#', 'Cis'),
		('Db', 'Des'),
	]
)


defaultLangNotation = {
	'en': enNotation,
	'it': itNotation,
	'fr': frNotation,
	'de': deNotation,
}


naturalScale = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']

tone = {
	'C': 0,
	'D': 2,
	'E': 4,
	'F': 5,
	'G': 7,
	'A': 9,
	'B': 11
}

interval = [
 (0, 0),
 (0, 1),
 (1, 0),
 (1, 1),
 (2, 0),
 (3, 0),
 (3, 1),
 (4, 0),
 (4, 1),
 (5, 0),
 (5, 1),
 (6, 0),
 (6, 1),
 (7, 0)
]

scales = {
	'C': (0, ['C', 'D', 'E', 'F', 'G', 'A', 'B']),
	'Db': (1, ['Db', 'Eb', 'F', 'Gb', 'Ab', 'Bb', 'C']),
	'D': (2, ['D', 'E', 'F#', 'G', 'A', 'B', 'C#']),
	'Eb': (3, ['Eb', 'F', 'G', 'Ab', 'Bb', 'C', 'D']),
	'E': (4, ['E', 'F#', 'G#', 'A', 'B', 'C#', 'D#']),
	'F': (5, ['F', 'G', 'A', 'Bb', 'C', 'D', 'E']),
	'Gb': (6, ['Gb', 'Ab', 'Bb', 'Cb', 'Db', 'Eb', 'F']),
	'G': (7, ['G', 'A', 'B', 'C', 'D', 'E', 'F#']),
	'Ab': (8, ['Ab', 'Bb', 'C', 'Db', 'Eb', 'F', 'G']),
	'A': (9, ['A', 'B', 'C#', 'D', 'E', 'F#', 'G#']),
	'Bb': (10, ['Bb', 'C', 'D', 'Eb', 'F', 'G', 'A']),
	'B': (11, ['B', 'C#', 'D#', 'E', 'F#', 'G#', 'A#'])
}
orderedKeys = ['C', 'Db', 'D', 'Eb', 'E', 'F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B']

vectorModes = ['', 'm', '7', 'm7']

referenceVector = [0.65713162540630443, 0.0, 0.037841466800806009, 0.0, 0.0, 0.0014554410308002313, 0.0, 0.0, 0.026197938554404162, 0.12444020813341977, 0.0014554410308002313, 0.029108820616004623, 0.0, 0.0, 0.0, 0.0, 0.040024628347006361, 0.12589564916422, 0.030564261646804855, 0.0036386025770005779, 0.46210252727907342, 0.013098969277202081, 0.0, 0.0, 0.0, 0.00072772051540011566, 0.0, 0.0, 0.49703311201827893, 0.0, 0.016009851338802544, 0.0, 0.00072772051540011566, 0.0, 0.0, 0.0, 0.0094603667002015022, 0.2648902676056421, 0.0058217641232009253, 0.0043663230924006939, 0.0014554410308002313, 0.0, 0.00072772051540011566, 0.0, 0.0, 0.0014554410308002313, 0.0, 0.0]


def splitChord(c, locNotation=enNotation):
	for k in locNotation.chords:
		if c.upper().startswith(k.upper()):
			if len(c) != len(k):
				d = c[len(k)]
				if d == "b" or d == "#":
					return (k+d, c[(len(k) + 1):])
				return (k, c[len(k):])
			return (c, "")
	return ("", c)


def __alteration(chord):
	if len(chord) == 1:
		return (chord, 0)
	elif chord[1] == '#':
		return (chord[0], 1)
	else:
		return (chord[0], -1)


def chord2pos(chord, key="C"):
	c, a = __alteration(chord)
	s, b = __alteration(key)
	return (tone[c.upper()] + a - tone[s.upper()] - b) % 12


def __pos2chord(pos, key):
	n, i = interval[pos]
	# if pos in scale, use it
	if i == 0:
		return scales[key][1][n]
	# else use natural scale
	ref = scales[key][0]
	diff = (pos + ref) % 12
	return naturalScale[diff]


def transpose(s, d, chord, notation=enNotation):
	sl = chord.find("/")
	if sl > -1:
		return "%s/%s" % (transpose(s, d, chord[:sl], notation), transpose(s, d, chord[sl + 1:], notation))
	chord = translateChord(chord, notation, enNotation)
	c, v = splitChord(chord)
	if c == "":
		return chord
	p = chord2pos(c, s)
	return translateChord(__pos2chord(p, d) + v, enNotation, notation)


def translateChord(chord, sNotation=enNotation, dNotation=enNotation):
	#if sNotation == dNotation:
	#	return chord
	sl = chord.find("/")
	if sl > -1:
		return "%s/%s" % (translateChord(chord[:sl], sNotation, dNotation), translateChord(chord[sl + 1:], sNotation, dNotation))
	chord = sNotation.PreprocessingToStandard(chord)
	c, a = splitChord(chord, sNotation)
	if c == "":
		return chord
	alt = c[-1]
	if alt == '#' or alt == 'b':
		c = c[:-1]
	else:
		alt = ""
	if c == "":
		return chord
	d = dNotation.Ord2Chord(sNotation.Chord2Ord(c)) + alt
	b = dNotation.AlterationFromStandard(sNotation.AlterationToStandard(a))
	d, b = dNotation.PostprocessingFromStandard(d, b)
	return d + b


def transposeChordPro(s, d, text, notation=enNotation):
	r = re.compile('\[([^]]*)\]')
	p = 0
	b = ''
	for m in r.finditer(text):
		b += "%s[%s]" % (
			text[p:m.start()],
			transpose(s, d, m.group(1), notation)
		)
		p = m.end()
	return b + text[p:]


def translateChordPro(text, sNotation=enNotation, dNotation=enNotation):
	r = re.compile('\[([^]]*)\]')
	p = 0
	b = ''
	for m in r.finditer(text):
		b += "%s[%s]" % (
			text[p:m.start()],
			translateChord(m.group(1), sNotation, dNotation)
		)
		p = m.end()
	return b + text[p:]


def autodetectNotation(text, notations):
	r = re.compile('\[([^]]*)\]')
	cnt = [0 for x in notations]
	for m in r.finditer(text):
		for i in range(0, len(notations)):
			c, a = splitChord(m.group(1), notations[i])
			if c != "":
				cnt[i] += 1
	return notations[cnt.index(max(cnt))]


def normalize(vector):
	count = 0
	for c in vector:
		count += c**2
	if count != 0:
		count = math.sqrt(count)
		return [x/count for x in vector]
	else:
		return vector


def scalarProduct(v1, v2):
	count = 0
	for i in range(0, len(v1)):
		count += v1[i]*v2[i]
	return count


def vectorizeChords(text, notation=enNotation):
	r = re.compile('\[([^]]*)\]')
	v = [0 for x in range(12 * len(vectorModes))]
	for m in r.finditer(text):
		chord = translateChord(m.group(1), notation, enNotation)
		c, a = splitChord(chord)
		if c != "" and a in vectorModes:
			v[chord2pos(c, "C") * len(vectorModes) + vectorModes.index(a)] += 1
	return normalize(v)


def autodetectKey(text, notation=enNotation):
	v = vectorizeChords(text, notation)
	r = referenceVector
	max = 0
	key = 0
	n = len(vectorModes)
	for k in range(0, 12):
		s = scalarProduct(v, r)
		if s > max:
			max = s
			key = k
		r = r[-n:] + r[:-n]
	return orderedKeys[key]


def integrateChords(chords, text):
	"""Integrate chord line in text line, as chordpro"""
	r = re.compile(r'(\S+)')
	i = r.finditer(chords)
	l = [x for x in i]
	l.reverse()
	if l != []:
		lc = l[0].start()
		lt = len(text)
		for i in range(0, lc-lt):
			text += ' '
		for m in l:
			s = m.start()
			text = text[:s] + '[' + m.group() + ']' + text[s:]
	return text


def testChordLine(line, notation=enNotation):
	"""Return True iff line contains only chords"""
	# First, tokenize line: if a token is not a chord => False
	r = re.compile(r'(\S+)')
	l = 0
	n = 0
	for m in r.finditer(line):
		c = m.group()
		c = notation.PreprocessingToStandard(c)
		e, a = splitChord(c, notation)
		if e == '':
			return False
		l += len(translateChord(c, notation, enNotation))
		n += 1
	# We don't consider an empty line a chord line
	if n == 0:
		return False
	# Looks okay... if a whitespace is occurring twice => True
	if line.find("  ") != -1:
		return True
	# Uhm, weird... are chord tokens "short"? Yes <=> True
	return l/float(n) <= 2.3


def testTabFormat(text, notations):
	"""
	Test whether text is in tab format

	It happens iff at least 3 lines are chord lines.
	Return the best matching notation, or None
	notations is a list of notations to be tested
	"""
	lines = text.splitlines()
	max = 0
	maxn = None
	for n in notations:
		nc = 0
		for l in lines:
			if testChordLine(l, n):
				nc += 1
		if nc > max and nc >= 3:
			max = nc
			maxn = n
	return maxn


def tab2ChordPro(text, notation=enNotation):
	"""
	Convert text from tab to chordpro format
	"""
	l = text.splitlines()
	n = len(l)
	i = 0
	out = []
	while i < n:
		if i + 1 < n and testChordLine(l[i], notation) and not testChordLine(l[i+1], notation):
			out.append(integrateChords(l[i], l[i+1]))
			i += 2
		else:
			out.append(l[i])
			i += 1
	return "\n".join(out)


def testSpuriousLines(text):
	"""
	Determine if there are spurious empty lines

	It happens iff at least 1/3 of the lines are empty,
	and there are at least 3 non-empty lines
	"""
	c = 0
	ll = text.splitlines()
	for l in ll:
		if l.strip() == '':
			c += 1
	return c >= len(ll)/3.0 and len(ll) - c >= 3


def removeSpuriousLines(text):
	"""
	Remove spurious blank lines from text

	If a blank line is isolated, remove it
	If there are several contiguous blank lines, keep only one of them
	"""
	out = []
	ll = text.splitlines()
	ln = len(ll)
	i = 0
	def empty(l):
		return l.strip() == ''
	while i < ln:
		if empty(ll[i]):
			if i + 1 < ln and empty(ll[i + 1]):
				out.append('')
				i += 2
				while i < ln and empty(ll[i]):
					i += 1
			else:
				i += 1
		else:
			out.append(ll[i])
			i += 1
	return "\n".join(out)


def findEasiestKey(text, fav, notation=enNotation):
	"""
	Find easiest key for song.

		text: song text
		fav: dictionary of (s)favourite chords, with the form {chord: weight}
		     (chords in fav are expressed using enNotation)
		return (chord_count, current_key, current_difficulty, easiest_key, easiest_difficulty)
	"""
	current_key = autodetectKey(text, notation)
	ws = dict([(k, 0) for k in scales])
	r = re.compile('\[([^]]*)\]')
	count = 0
	for m in r.finditer(text):
		count += 1
		chord = translateChord(m.group(1), notation)
		for k in ws:
			c = transpose(current_key, k, chord)
			if c in fav:
				ws[k] += fav[c]
	easiest_key = current_key
	m = ws[easiest_key]
	for k in ws:
		if m is None or ws[k] > m:
			easiest_key = k
			m = ws[k]
	difficulty = lambda x: max(0, min(1, (count - x) / float(count)))
	if count > 0:
		return (
				count,
				translateChord(current_key, enNotation, notation),
				difficulty(ws[current_key]),
				translateChord(easiest_key, enNotation, notation),
				difficulty(m),
			)
	return (0, current_key, 0, current_key, 0)


def removeChords(text):
	"""
	Remove all chords in (ChordPro) text.

		text: song text
		return: text without chords
	"""
	return re.sub('\[([^]]*)\]', "", text)


def removeChordPro(text):
	"""
	Remove every ChordPro tag, including chords, from text.

		text: song text
		return: text without ChordPro stuff and chords
	"""
	return removeSpuriousLines(re.sub('(\[([^]]*)\]|{[^}]*})', "", text))


def replaceTitles(text, replace_string):
	"""
	Replace {t:something} and {tititle:something} commands with a string

	:param text: ChordPro song
	:param replace_string: string to be replaced to titles
	:return:
	"""
	return re.sub(r'({t:[^}]*}|{title:[^}]*})', replace_string, text)


def pasteChords(src, dest):
	"""
	Paste chords from src to dest

	Remove any existing chords in lines in the first k lines of dest,
	if there are k lines in src.
		src: source chordpro text
		dest: destination text
	"""
	ss = src.splitlines()
	sd = dest.splitlines()
	out = []
	m = min(len(ss), len(sd))
	r = re.compile('\[([^]]*)\]')
	for i in range(0, m):
		cd = r.sub("", sd[i])
		for x in r.finditer(ss[i]):
			s = x.start()
			if len(cd) < s:
				cd += "".join([" " for y in range(len(cd), s)])
			cd = cd[:s] + x.group(0) + cd[s:]
		out.append(cd)
	for i in range(m, len(sd)):
		out.append(r.sub("", sd[i]))
	return "\n".join(out)



# -*- coding: utf-8 -*- 

###########################################################################
## Python code generated with wxFormBuilder (version Sep  8 2010)
## http://www.wxformbuilder.org/
##
## PLEASE DO "NOT" EDIT THIS FILE!
###########################################################################

import wx

import gettext
_ = gettext.gettext

###########################################################################
## Class CompositePropertyPanel
###########################################################################

class CompositePropertyPanel ( wx.Panel ):
	
	def __init__( self, parent ):
		wx.Panel.__init__ ( self, parent, id = wx.ID_ANY, pos = wx.DefaultPosition, size = wx.Size( -1,-1 ), style = wx.TAB_TRAVERSAL )
		
		bSizer21 = wx.BoxSizer( wx.VERTICAL )
		
		bSizer24 = wx.BoxSizer( wx.HORIZONTAL )
		
		self.m_bitmap1 = wx.StaticBitmap( self, wx.ID_ANY, wx.NullBitmap, wx.DefaultPosition, wx.DefaultSize, 0 )
		bSizer24.Add( self.m_bitmap1, 0, wx.ALL, 5 )
		
		self.label = wx.StaticText( self, wx.ID_ANY, _("label"), wx.DefaultPosition, wx.DefaultSize, 0 )
		self.label.Wrap( -1 )
		bSizer24.Add( self.label, 0, wx.ALIGN_CENTER|wx.ALL, 5 )
		
		self.check = wx.CheckBox( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
		self.check.SetValue(True) 
		bSizer24.Add( self.check, 0, wx.ALIGN_CENTER|wx.ALL, 5 )
		
		bSizer21.Add( bSizer24, 0, wx.EXPAND, 5 )
		
		"""
		self.value = self.get_widget()
		bSizer21.Add( self.value, 1, wx.ALIGN_CENTER|wx.BOTTOM|wx.EXPAND|wx.LEFT|wx.TOP, 5 )
		"""
		
		self.SetSizer( bSizer21 )
		self.Layout()
		bSizer21.Fit( self )
		
		# Connect Events
		self.check.Bind( wx.EVT_CHECKBOX, self.OnCheck )
	
	def __del__( self ):
		pass
	
	
	# Virtual event handlers, overide them in your derived class
	def OnCheck( self, event ):
		event.Skip()
	




###############################################################
# Name:			 PreferencesDialog.py
# Purpose:	 Allow user to set preferences
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-10-02
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

import wx

from .PreferencesDialog import PreferencesDialog
from .MyDecoSlider import MyDecoSlider
from .Transpose import *
from .Preferences import get_update_frequencies
from .Globals import glb


i18n.register('songpress.MyPreferencesDialog')


class MyPreferencesDialog(PreferencesDialog):
	def __init__(self, parent, preferences, easyChords):
		self.pref = preferences
		self.frame = self
		PreferencesDialog.__init__(self, parent)

		self.easyChords = easyChords

		self.fontCB.Bind(wx.EVT_TEXT_ENTER, self.OnFontSelected, self.fontCB)
		self.fontCB.Bind(wx.EVT_COMBOBOX, self.OnFontSelected, self.fontCB)

		previewSong = _("{t:My Bonnie}\n\nMy [D]Bonnie lies [G]over the [D]ocean\noh [G]bring back my [A]Bonnie to [D]me!\n\n{soc}\n[D]Bring back, [E-]bring back,\n[A]bring back my Bonnie to [D]me!\n{eoc}")
		self.editor.SetText(previewSong)
		self.editor.SetFont(self.pref.editorFace, self.pref.editorSize)
		self.editor.SetReadOnly(True)
		self.autoRemoveBlankLines.SetValue(self.pref.autoAdjustSpuriousLines)
		self.autoTab2Chordpro.SetValue(self.pref.autoAdjustTab2Chordpro)
		self.autoAdjustEasyKey.SetValue(self.pref.autoAdjustEasyKey)
		if self.pref.locale is None:
			lang = i18n.getLang()
		else:
			lang = self.pref.locale
		for l in glb.languages:
			i = self.langCh.Append(glb.languages[l])
			self.langCh.SetClientData(i, l)
			if lang == l:
				self.langCh.SetSelection(i)
		exts = ["crd", "pro", "chopro", "chordpro", "cho"]
		i = 0
		for e in exts:
			self.extension.Append(e)
			if e == self.pref.defaultExtension:
				self.extension.SetSelection(i)
			i += 1
				
		# Default notation
		for n in self.pref.notations:
			i = self.notationCh.Append(n.desc)
			self.notationCh.SetClientData(i, n.id)
		self.notationCh.SetSelection(0)		

		# Update frequency
		sel = 0
		uf = get_update_frequencies()
		for k in uf:
			i = self.frequency.Append(uf[k])
			self.frequency.SetClientData(i, k)
			if k == self.pref.updateFrequency:
				sel = i
		self.frequency.SetSelection(sel)

		# Easy chords
		simplifyGrid = wx.FlexGridSizer(len(easyChords), 2, 0, 0)
		simplifyGrid.AddGrowableCol(1, 1)
		self.simplifyPanel.SetSizer(simplifyGrid)
		self.simplifyPanel.Layout()

		self.decoSliders = {}
		for k in easyChordsOrder:
			simplifyGrid.Add(wx.StaticText(self.simplifyPanel, wx.ID_ANY, getEasyChordsDescription(easyChords[k]), wx.DefaultPosition, wx.DefaultSize, 0), 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
			ds = MyDecoSlider(self.simplifyPanel)
			self.decoSliders[k] = ds
			ds.slider.SetValue(self.pref.GetEasyChordsGroup(k))
			simplifyGrid.Add(ds, 1, wx.EXPAND, 5)

		simplifyGrid.FitInside(self.simplifyPanel)

		#self.simplifyPanel.Refresh()

	def OnFontSelected(self, evt):
		f, s = self.GetFont()
		self.editor.SetFont(f, s)
		evt.Skip()

	def GetFont(self):
		face = self.fontCB.GetValue()
		try:
			s = int(self.sizeCB.GetValue())
		except:
			s = 12
		return (face, s)

	def GetLanguage(self):
		return self.langCh.GetClientData(self.langCh.GetSelection())

	def GetNotation(self):
		return self.notationCh.GetClientData(self.notationCh.GetSelection())

	def GetUpdateFrequency(self):
		return self.frequency.GetClientData(self.frequency.GetSelection())

	def OnOk(self, evt):
		self.pref.editorFace, self.pref.editorSize = self.GetFont()
		self.pref.updateFrequency = self.GetUpdateFrequency()
		l = self.GetLanguage()
		lang = i18n.getLang()
		if l is not None and l != lang:
			msg = _("Language settings will be applied when you restart Songpress.")
			d = wx.MessageDialog(self, msg, _("Songpress"), wx.ICON_INFORMATION | wx.OK)
			d.ShowModal()
		self.pref.locale = l
		self.pref.SetDefaultNotation(self.GetNotation())
		self.pref.autoAdjustSpuriousLines = self.autoRemoveBlankLines.GetValue()
		self.pref.autoAdjustTab2Chordpro = self.autoTab2Chordpro.GetValue()
		self.pref.autoAdjustEasyKey = self.autoAdjustEasyKey.GetValue()
		for k in self.decoSliders:
			self.pref.SetEasyChordsGroup(k, self.decoSliders[k].slider.GetValue())
		self.pref.defaultExtension = self.extension.GetString(self.extension.GetSelection())
		evt.Skip(True)




# -*- coding: utf-8 -*- 

###########################################################################
## Python code generated with wxFormBuilder (version Aug 25 2009)
## http://www.wxformbuilder.org/
##
## PLEASE DO "NOT" EDIT THIS FILE!
###########################################################################

import wx

from . import i18n


i18n.register('songpress.TransposeDialog')

###########################################################################
## Class TransposeDialog
###########################################################################

class TransposeDialog ( wx.Dialog ):
	
	def __init__( self, parent ):
		wx.Dialog.__init__  ( self, parent, id = wx.ID_ANY, title = _(u"Transpose"), pos = wx.DefaultPosition, size = wx.Size( -1,-1 ), style = wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER )
		
		self.SetSizeHintsSz( wx.DefaultSize, wx.DefaultSize )
		
		bSizer1 = wx.BoxSizer( wx.VERTICAL )
		
		bSizer2 = wx.BoxSizer( wx.HORIZONTAL )
		
		self.label1 = wx.StaticText( self, wx.ID_ANY, _(u"Chord notation:"), wx.DefaultPosition, wx.DefaultSize, 0 )
		self.label1.Wrap( -1 )
		bSizer2.Add( self.label1, 0, wx.ALL, 5 )
		
		notationChoices = []
		self.notation = wx.Choice( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, notationChoices, 0 )
		self.notation.SetSelection( 0 )
		bSizer2.Add( self.notation, 1, wx.ALL, 5 )
		
		bSizer1.Add( bSizer2, 1, wx.EXPAND, 5 )
		
		bSizer21 = wx.BoxSizer( wx.HORIZONTAL )
		
		self.label11 = wx.StaticText( self, wx.ID_ANY, _(u"From key:"), wx.DefaultPosition, wx.DefaultSize, 0 )
		self.label11.Wrap( -1 )
		bSizer21.Add( self.label11, 0, wx.ALL, 5 )
		
		fromKeyChoices = []
		self.fromKey = wx.Choice( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, fromKeyChoices, 0 )
		self.fromKey.SetSelection( 0 )
		bSizer21.Add( self.fromKey, 1, wx.ALL, 5 )
		
		bSizer1.Add( bSizer21, 1, wx.EXPAND, 5 )
		
		bSizer2111 = wx.BoxSizer( wx.HORIZONTAL )
		
		self.label1111 = wx.StaticText( self, wx.ID_ANY, _(u"Semitones:"), wx.DefaultPosition, wx.DefaultSize, 0 )
		self.label1111.Wrap( -1 )
		bSizer2111.Add( self.label1111, 0, wx.ALL, 5 )
		
		self.semitones = wx.SpinCtrl( self, wx.ID_ANY, u"2", wx.DefaultPosition, wx.DefaultSize, wx.SP_ARROW_KEYS|wx.SP_WRAP, -11, 12, 2 )
		bSizer2111.Add( self.semitones, 1, wx.ALL, 5 )
		
		bSizer1.Add( bSizer2111, 1, wx.EXPAND, 5 )
		
		bSizer211 = wx.BoxSizer( wx.HORIZONTAL )
		
		self.label111 = wx.StaticText( self, wx.ID_ANY, _(u"To key:"), wx.DefaultPosition, wx.DefaultSize, 0 )
		self.label111.Wrap( -1 )
		bSizer211.Add( self.label111, 0, wx.ALL, 5 )
		
		toKeyChoices = []
		self.toKey = wx.Choice( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, toKeyChoices, 0 )
		self.toKey.SetSelection( 0 )
		bSizer211.Add( self.toKey, 1, wx.ALL, 5 )
		
		bSizer1.Add( bSizer211, 1, wx.EXPAND, 5 )
		
		m_sdbSizer1 = wx.StdDialogButtonSizer()
		self.m_sdbSizer1OK = wx.Button( self, wx.ID_OK )
		m_sdbSizer1.AddButton( self.m_sdbSizer1OK )
		self.m_sdbSizer1Cancel = wx.Button( self, wx.ID_CANCEL )
		m_sdbSizer1.AddButton( self.m_sdbSizer1Cancel )
		m_sdbSizer1.Realize();
		bSizer1.Add( m_sdbSizer1, 0, wx.ALL|wx.EXPAND, 5 )
		
		self.SetSizer( bSizer1 )
		self.Layout()
		bSizer1.Fit( self )
		
		# Connect Events
		self.notation.Bind( wx.EVT_CHOICE, self.OnNotation )
		self.fromKey.Bind( wx.EVT_CHOICE, self.OnFromKey )
		self.semitones.Bind( wx.EVT_SPINCTRL, self.OnSemitones )
		self.toKey.Bind( wx.EVT_CHOICE, self.OnToKey )
	
	def __del__( self ):
		pass
	
	
	# Virtual event handlers, overide them in your derived class
	def OnNotation( self, event ):
		event.Skip()
	
	def OnFromKey( self, event ):
		event.Skip()
	
	def OnSemitones( self, event ):
		event.Skip()
	
	def OnToKey( self, event ):
		event.Skip()
	




###############################################################
# Name:			 FontComboBox.py
# Purpose:	 wxWidgets control providing a combo box font
#            selector
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-08-13
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

import wx

class MyFontEnumerator(wx.FontEnumerator):
	def OnFacename(self, name):
		if name[0] != '@':
			self.names.append(name)
		return True
		
	def Enumerate(self):
		self.names = []
		self.EnumerateFacenames()
		return self
		
	def AppendToComboBox(self, cb):
		self.names.sort(key=lambda x: x.upper())
		cb.AppendItems(self.names)

class FontComboBox(wx.ComboBox):
	def __init__(self, parent, id=-1, defaultFont="Arial"):
		s = wx.Size()
		s.width = 200
		s.height = -1
		wx.ComboBox.__init__(self, parent, id, defaultFont, size=s)
		MyFontEnumerator().Enumerate().AppendToComboBox(self)



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version Mar 10 2018)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
import wx.xrc

# Nothing to localize in this module, because all strings are "dummy" placeholders to be replaced

_ = lambda x: x


###########################################################################
## Class ListDialog
###########################################################################

class ListDialog(wx.Dialog):

	def __init__(self, parent):
		wx.Dialog.__init__(self, parent, id=wx.ID_ANY, title=_(u"Songpress"), pos=wx.DefaultPosition,
											 size=wx.Size(413, 240), style=wx.DEFAULT_DIALOG_STYLE)

		self.SetSizeHints(wx.DefaultSize, wx.DefaultSize)

		bSizer24 = wx.BoxSizer(wx.VERTICAL)

		self.label = wx.StaticText(self, wx.ID_ANY, _(u"Label text"), wx.DefaultPosition, wx.DefaultSize, 0)
		self.label.Wrap(-1)
		bSizer24.Add(self.label, 0, wx.ALL, 5)

		elementsChoices = []
		self.elements = wx.ListBox(self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, elementsChoices, wx.LB_SINGLE)
		bSizer24.Add(self.elements, 3, wx.ALL | wx.EXPAND, 5)

		m_sdbSizer5 = wx.StdDialogButtonSizer()
		self.m_sdbSizer5OK = wx.Button(self, wx.ID_OK)
		m_sdbSizer5.AddButton(self.m_sdbSizer5OK)
		self.m_sdbSizer5Cancel = wx.Button(self, wx.ID_CANCEL)
		m_sdbSizer5.AddButton(self.m_sdbSizer5Cancel)
		m_sdbSizer5.Realize();

		bSizer24.Add(m_sdbSizer5, 0, wx.EXPAND, 5)

		self.SetSizer(bSizer24)
		self.Layout()

		self.Centre(wx.BOTH)

	def __del__(self):
		pass





# -*- coding: utf-8 -*- 

###########################################################################
## Python code generated with wxFormBuilder (version Sep  8 2010)
## http://www.wxformbuilder.org/
##
## PLEASE DO "NOT" EDIT THIS FILE!
###########################################################################

import wx

import gettext
_ = gettext.gettext

###########################################################################
## Class DecoSlider
###########################################################################

class DecoSlider ( wx.Panel ):
	
	def __init__( self, parent ):
		wx.Panel.__init__ ( self, parent, id = wx.ID_ANY, pos = wx.DefaultPosition, size = wx.Size( -1,-1 ), style = wx.TAB_TRAVERSAL )
		
		self.SetBackgroundColour( wx.SystemSettings.GetColour( wx.SYS_COLOUR_WINDOW ) )
		
		bSizer9 = wx.BoxSizer( wx.VERTICAL )
		
		self.slider = wx.Slider( self, wx.ID_ANY, 0, 0, 4, wx.DefaultPosition, wx.DefaultSize, wx.SL_AUTOTICKS|wx.SL_HORIZONTAL )
		bSizer9.Add( self.slider, 0, wx.ALL|wx.EXPAND, 5 )
		
		self.panel = wx.Panel( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL )
		self.panel.SetBackgroundColour( wx.SystemSettings.GetColour( wx.SYS_COLOUR_WINDOW ) )
		self.panel.SetMinSize( wx.Size( -1,30 ) )
		
		bSizer9.Add( self.panel, 1, wx.EXPAND |wx.ALL, 5 )
		
		self.SetSizer( bSizer9 )
		self.Layout()
		bSizer9.Fit( self )
		
		# Connect Events
		self.panel.Bind( wx.EVT_PAINT, self.OnPaint )
		self.panel.Bind( wx.EVT_SIZE, self.OnSize )
	
	def __del__( self ):
		pass
	
	
	# Virtual event handlers, overide them in your derived class
	def OnPaint( self, event ):
		event.Skip()
	
	def OnSize( self, event ):
		event.Skip()
	




###############################################################
# Name:			 Renderer.py
# Purpose:	 Render a song on a dc
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-02-21
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

import wx

from .SongDecorator import *
from .SongTokenizer import *
from .SongFormat import *
from .SongBoxes import *
from .Transpose import translateChord, autodetectNotation
from .EditDistance import minEditDist


class BreakException(Exception):
	pass


class Renderer(object):
	def __init__(self, sf, sd=SongDecorator(), notations=[]):
		object.__init__(self)
		self.text = ""
		self.sd = sd
		self.dc = None
		# SongFormat
		self.starting_sf = sf
		self.sf = sf
		self.format = None
		self.currentBlock = None
		self.currentLine = None
		self.song = None
		self.notation = None
		self.notations = notations
		self.chordPatterns = []

	def BeginBlock(self, type, label=None):
		self.EndBlock()
		if type == SongBlock.verse:
			self.song.verseCount += 1
			if label is None:
				self.song.labelCount += 1
			self.sf.AddVerse()
			self.format = self.sf.verse[-1]
		elif type == SongBlock.chorus:
			self.format = self.sf.chorus
			self.song.chorusCount += 1
		else:
			self.format = self.sf.title
		self.currentBlock = SongBlock(type, self.format)
		self.currentBlock.label = label
		self.currentBlock.verseNumber = self.song.verseCount
		self.currentBlock.verseLabelNumber = self.song.labelCount

	def EndBlock(self):
		if self.currentBlock is not None:
			self.EndLine()
			if self.sf.showChords == 1:
				current = self.currentBlock.chords
				found = False
				for p in self.chordPatterns:
					led = len(p)
					med = minEditDist(p, current)
					if med < led and med < 4:
						found = True
						break
				if not found:
					self.chordPatterns.append(current)
				else:
					self.currentBlock.RemoveChordBoxes()
			self.song.AddBox(self.currentBlock)
			self.currentBlock = None

	def BeginVerse(self, label=None):
		if self.currentBlock == None:
			self.BeginBlock(SongBlock.verse, label)
			self.label = label

	def BeginChorus(self, label=None):
		self.BeginBlock(SongBlock.chorus, label)

	def ChorusVSkip(self):
		self.EndLine()
		self.BeginLine()
		self.EndLine()

	def AddText(self, text, type=SongText.text):
		if(
			text.strip() != ''
			and type != SongText.title
			and type != SongText.subtitle
			and type != SongText.comment
			and self.currentBlock is not None
			and self.currentBlock.type == SongBlock.title
		):
			self.EndBlock()
		self.BeginVerse()
		self.BeginLine()
		if type == SongText.comment:
			text = "(" + text + ")"
			format = self.format.comment
		elif type == SongText.chord:
			format = self.format.chord
			if self.sf.showChords == 1:
				self.currentBlock.chords.append(translateChord(text, self.notation, self.notation))
		else:
			format = self.format
		t = SongText(text, format.wxFont, type, format.color)
		if not type == SongText.chord or self.sf.showChords > 0:
			self.currentLine.AddBox(t)

	def AddTitle(self, title):
		if self.currentBlock is None or self.currentBlock.type != SongBlock.title:
			self.BeginBlock(SongBlock.title)
		self.format = self.sf.title
		self.AddText(title, SongText.title)

	def AddSubTitle(self, title):
		if self.currentBlock is None or self.currentBlock.type != SongBlock.title:
			self.BeginBlock(SongBlock.title)
		self.format = self.sf.subtitle
		self.AddText(title, SongText.subtitle)

	def BeginLine(self):
		if self.currentLine == None:
			if self.song.drawWholeSong or (self.fromLine <= self.lineCount and self.lineCount <= self.toLine):
				self.currentBlock.drawBlock = True
			self.currentLine = SongLine()

	def EndLine(self):
		if self.currentLine is not None:
			self.currentBlock.AddBox(self.currentLine)
			self.currentLine = None

	def GetAttribute(self):
		try:
			tok = self.tkz.next()
			if tok.token != SongTokenizer.colonToken:
				self.tkz.Repeat()
				return None
			tok = self.tkz.next()
			if tok.token != SongTokenizer.attrToken:
				self.tkz.Repeat()
				return ''
			return tok.content
		except StopIteration:
			pass
		return None

	def GetState(self):
		return None if self.currentBlock == None else self.currentBlock.type

	def Render(self, text, dc, fromLine = -1, toLine = -1):
		self.text = text
		self.dc = dc
		self.verseNumber = 0
		self.format = self.starting_sf
		self.currentLine = None
		self.currentBlock = None
		self.song = SongSong(self.sf)
		self.song.drawWholeSong = fromLine == -1
		self.lineCount = -1
		self.fromLine = fromLine
		self.toLine = toLine
		self.sf = SongFormat(self.starting_sf)
		if self.sf.showChords == 1:
			self.notation = autodetectNotation(text, self.notations)
			self.chordPatterns = []

		for l in self.text.splitlines():
			self.lineCount += 1
			state = self.GetState()
			self.tkz = SongTokenizer(l)
			empty = True
			for tok in self.tkz:
				state = self.GetState()
				empty = False
				t = tok.token
				if t == SongTokenizer.normalToken:
					self.AddText(tok.content)
				elif t == SongTokenizer.chordToken:
					self.AddText(tok.content[1:], SongText.chord)
				elif t == SongTokenizer.commandToken:
					cmd = tok.content.lower()
					if cmd == 'soc' or cmd == 'start_of_chorus':
						a = self.GetAttribute()
						self.BeginChorus(a)
					elif (cmd == 'eoc' or cmd == 'end_of_chorus') and state == SongBlock.chorus:
						self.EndBlock()
					elif cmd == 'c' or cmd == 'comment':
						a = self.GetAttribute()
						if a is not None:
							self.AddText(a, SongText.comment)
					elif cmd == 't' or cmd == 'title':
						a = self.GetAttribute()
						if a is not None:
							self.AddTitle(a)
					elif cmd == 'st' or cmd == 'subtitle':
						a = self.GetAttribute()
						if a is not None:
							self.AddSubTitle(a)
					elif cmd == 'verse':
						self.BeginVerse(self.GetAttribute())
					elif cmd == 'textsize':
						try:
							a = self.GetAttribute()
							if a is None:
								size = self.starting_sf.size
							else:
								a = a.strip()
								try:
									if a.endswith("%"):
										perc = int(a[:-1])
										size = self.format.size * perc / 100
									else:
										size = int(a)
								except TypeError:
									raise BreakException()
								except ValueError:
									raise BreakException()
							self.format = ParagraphFormat(self.format)
							self.format.size = size
							self.sf.size = size
							self.sf.chorus.size = size
							self.sf.title.size = size
							self.sf.subtitle.size = size
						except BreakException:
							pass
					elif cmd == 'textfont':
						try:
							face = self.GetAttribute()
							if face is None:
								face = self.starting_sf.face
							else:
								face = face.strip()
							self.format = ParagraphFormat(self.format)
							self.format.face = face
							self.sf.face = face
							self.sf.chorus.face = face
							self.sf.title.face = face
							self.sf.subtitle.face = face
						except BreakException:
							pass
					elif cmd == 'textcolour':
						try:
							color = self.GetAttribute()
							if color is None:
								color = self.starting_sf.color
							else:
								color = color.strip()
							self.format = ParagraphFormat(self.format)
							self.format.color = color
							self.sf.color = color
							self.sf.chorus.color = color
							self.sf.title.color = color
							self.sf.subtitle.color = color
						except BreakException:
							pass
					elif cmd == 'chordsize':
						try:
							a = self.GetAttribute()
							if a is None:
								size = self.starting_sf.chord.size
							else:
								a = a.strip()
								try:
									if a.endswith("%"):
										perc = int(a[:-1])
										size = self.format.chord.size * perc / 100
									else:
										size = int(a)
								except TypeError:
									raise BreakException()
								except ValueError:
									raise BreakException()
							self.format = ParagraphFormat(self.format)
							self.format.chord.size = size
							self.sf.chord.size = size
							self.sf.chorus.chord.size = size
						except BreakException:
							pass
					elif cmd == 'chordfont':
						try:
							face = self.GetAttribute()
							if face is None:
								face = self.starting_sf.chord.face
							else:
								face = face.strip()
							self.format = ParagraphFormat(self.format)
							self.format.chord.face = face
							self.sf.chord.face = face
							self.sf.chorus.chord.face = face
						except BreakException:
							pass
					elif cmd == 'chordcolour':
						try:
							color = self.GetAttribute()
							if color is None:
								color = self.starting_sf.chord.color
							else:
								color = color.strip()
							self.format = ParagraphFormat(self.format)
							self.format.chord.color = color
							self.sf.chord.color = color
							self.sf.chorus.chord.color = color
						except BreakException:
							pass

			self.EndLine()
			if empty:
				if state in {SongBlock.verse, SongBlock.title}:
					self.EndBlock()
				elif state == SongBlock.chorus:
					self.ChorusVSkip()
		self.EndBlock()
		w, h = self.sd.Draw(self.song, dc)
		self.dc = None
		return w, h

	def SetDecorator(self, sd):
		self.sd = sd



#!/usr/bin/env python

import wx

from .UpdatePanel import UpdatePanel


class MyUpdatePanel(UpdatePanel):
	def __init__(self, parent, preferences, version, description, downloadUrl):
		UpdatePanel.__init__(self, parent)
		self.parent = parent
		self.preferences = preferences
		self.downloadUrl = downloadUrl
		self.v = version
		self.version.SetValue(version)
		self.new_features.SetPage(description)
		
	def __Hide(self):
		self.Show(False)
		self.parent.RemoveChild(self)
		
	def OnDownload(self, evt):
		wx.LaunchDefaultBrowser(self.downloadUrl)
		self.__Hide()
		evt.Skip()
		
	def OnSkip(self, evt):
		self.preferences.ignoredUpdates.add(self.v)
		self.__Hide()
		evt.Skip()
		
	def OnRemind(self, evt):
		self.__Hide()
		evt.Skip()




###############################################################
# Name:			 SongpressFrame.py
# Purpose:	 Main frame for Songpress
# Author:		 Luca Allulli (webmaster@roma21.it)
# Created:	 2009-01-16
# Copyright: Luca Allulli (https://www.skeed.it/songpress)
# License:	 GNU GPL v2
##############################################################

import subprocess
import os
import os.path
import platform

# import wx.aui as aui
import wx.adv
from wx import xrc

from .Editor import *
from .PreviewCanvas import *
from .Renderer import *
from .FontComboBox import FontComboBox
from .FontFaceDialog import FontFaceDialog
from .MyPreferencesDialog import MyPreferencesDialog
from .HTML import HtmlExporter, TabExporter
from .MyTransposeDialog import *
from .MyNotationDialog import *
from .MyNormalizeDialog import *
from .MyListDialog import MyListDialog
from . import MyUpdateDialog
from .Globals import glb
from .Preferences import Preferences
from . import i18n

if platform.system() == 'Windows':
	import wx.msw

i18n.register('songpress.SongpressFrame')


class SongpressFindReplaceDialog(object):
	def __init__(self, owner, replace=False):
		object.__init__(self)
		self.down = True
		self.st = ''
		self.data = wx.FindReplaceData(wx.FR_DOWN)
		self.owner = owner
		self.flags = 0
		self.dialog = wx.FindReplaceDialog(
			owner.frame,
			self.data,
			_("Replace") if replace else _("Find"),
			wx.FR_REPLACEDIALOG if replace else 0
		)
		owner.frame.Bind(wx.EVT_FIND, self.OnFind, self.dialog)
		owner.frame.Bind(wx.EVT_FIND_NEXT, self.OnFind, self.dialog)
		owner.frame.Bind(wx.EVT_FIND_CLOSE, self.OnClose, id=wx.ID_ANY)
		owner.frame.Bind(wx.EVT_FIND_REPLACE, self.OnReplace, self.dialog)
		owner.frame.Bind(wx.EVT_FIND_REPLACE_ALL, self.OnReplaceAll, self.dialog)
		self.dialog.Show()

	def OnFind(self, evt):
		self.st = self.data.GetFindString()
		f = self.data.GetFlags()
		self.down = f & wx.FR_DOWN
		self.whole = f & wx.FR_WHOLEWORD
		self.case = f & wx.FR_MATCHCASE
		self.flags = 0
		if self.whole:
			self.flags |= wx.stc.STC_FIND_WHOLEWORD
		if self.case:
			self.flags |= wx.stc.STC_FIND_MATCHCASE
		self.FindNext()

	def FindNext(self):
		if self.down:
			self.search = self.owner.text.SearchNext
		else:
			self.search = self.owner.text.SearchPrev
		s, e = self.owner.text.GetSelection()
		if self.down:
			self.owner.text.SetSelection(e, e)
			fromStart = s == 0
		else:
			self.owner.text.SetSelection(s, s)
			fromStart = s == self.owner.text.GetLength()
		self.owner.text.SearchAnchor()
		p = self.search(self.flags, self.st)
		if p != -1:
			pass
			self.owner.text.SetSelection(p, p + len(self.st))
		else:
			parent = self.dialog if self.dialog != None else self.owner.frame
			if not fromStart:
				if self.down:
					where = _("beginning")
					wherefrom = _("Reached the end")
					newStart = 0
				else:
					where = _("end")
					wherefrom = _("Reached the beginning")
					newStart = self.owner.text.GetLength()
				d = wx.MessageDialog(
					parent,
					_("%s of the song, restarting search from the %s") % (wherefrom, where,),
					self.owner.appName,
					wx.OK | wx.CANCEL | wx.ICON_INFORMATION
				)
				res = d.ShowModal()
				if res == wx.ID_OK:
					self.owner.text.SetSelection(newStart, newStart)
					self.FindNext()
			else:
				d = wx.MessageDialog(
					parent,
					_("The specified text was not found"),
					self.owner.appName,
					wx.OK | wx.ICON_INFORMATION
				)
				res = d.ShowModal()

	def OnReplace(self, evt):
		r = self.data.GetReplaceString()
		self.st = self.data.GetFindString()
		if self.owner.text.GetSelectedText().lower() == self.st.lower():
			self.owner.text.ReplaceSelection(r)
			self.FindNext()

	def OnReplaceAll(self, evt):
		self.owner.text.BeginUndoAction()
		s = self.data.GetFindString()
		r = self.data.GetReplaceString()
		f = self.data.GetFlags()
		self.whole = f & wx.FR_WHOLEWORD
		self.case = f & wx.FR_MATCHCASE
		flags = 0
		if self.whole:
			flags |= wx.stc.STC_FIND_WHOLEWORD
		if self.case:
			flags |= wx.stc.STC_FIND_MATCHCASE
		self.owner.text.SetSelection(0, 0)
		c = 0
		p = 0
		while (p := self.owner.text.FindText(p, self.owner.text.GetLength(), s, flags)[0]) != -1:
			self.owner.text.SetTargetStart(p)
			self.owner.text.SetTargetEnd(p + len(s))
			p += self.owner.text.ReplaceTarget(r)
			c += 1

		self.owner.text.EndUndoAction()

		d = wx.MessageDialog(
			self.dialog,
			_("%d text occurrences have been replaced") % (c,),
			self.owner.appName,
			wx.OK | wx.ICON_INFORMATION
		)
		res = d.ShowModal()

	def OnClose(self, evt):
		self.dialog.Destroy()
		self.dialog = None


if platform.system() == 'Linux':
    # Apparently there is a problem with linux FileOpen dialog box in wxPython:
    # it does not support multiple extensions in a filter.
    _import_formats = [
        (_("All supported files"), ["crd", "cho", "chordpro", "chopro", "tab", "cpm"]),
        #(_("Chordpro files (*.crd)"), ["crd"]),
        #(_("Tab files (*.tab)"), ["tab"]),
        #(_("Chordpro files (*.cho)"), ["cho"]),
        #(_("Chordpro files (*.chordpro)"), ["chordpro"]),
        #(_("Chordpro files (*.chopro)"), ["chopro"]),
        #(_("Chordpro files (*.pro)"), ["pro"]),
    ]
else:
    _import_formats = [
        (_("All supported files"), ["crd", "cho", "chordpro", "chopro", "pro", "tab"]),
        (_("Chordpro files (*.crd, *.cho, *.chordpro, *.chopro, *.pro)"), ["crd", "cho", "chordpro", "chopro", "pro"]),
        (_("Tab files (*.tab)"), ["tab"]),
    ]


class SongpressFrame(SDIMainFrame):
	def __init__(self, res):
		SDIMainFrame.__init__(
			self,
			res,
			'MainFrame',
			'songpress',
			'Skeed',
			_('song'),
			'crd',
			_('Songpress - Il Canzonatore'),
			glb.AddPath('img/songpress.ico'),
			glb.VERSION,
			_("http://www.skeed.it/songpress"),
			(_(u"Copyright (c) 2009-{year} Luca Allulli - Skeed\nLocalization:\n{translations}")).format(
				year=glb.YEAR,
				translations="\n".join([u"- {}: {}".format(glb.languages[x], glb.translators[x]) for x in glb.languages])
			),
			_("Licensed under the terms and conditions of the GNU General Public License, version 2"),
			_(
				"Special thanks to:\n  * The Pyhton programming language (http://www.python.org)\n  * wxWidgets (http://www.wxwidgets.org)\n  * wxPython (http://www.wxpython.org)\n  * Editra (http://editra.org/) (for the error reporting dialog and... the editor itself!)\n  * python-pptx (for PowerPoint export)"),
			_import_formats,
		)
		self.pref = Preferences()
		self.SetDefaultExtension(self.pref.defaultExtension)
		self.text = Editor(self)
		dt = SDIDropTarget(self)
		self.text.SetDropTarget(dt)
		self.frame.Bind(wx.stc.EVT_STC_UPDATEUI, self.OnUpdateUI, self.text)
		# Other objects
		self.previewCanvas = PreviewCanvas(self.frame, self.pref.format, self.pref.notations, self.pref.decorator)
		self.AddMainPane(self.text)
		self.AddPane(self.previewCanvas.main_panel, aui.AuiPaneInfo().Right().BestSize(240, 400), _('Preview'), 'preview')
		if self.previewCanvas.link is not None:
			self.previewCanvas.main_panel.Bind(wx.adv.EVT_HYPERLINK, self.OnCopyAsImage, self.previewCanvas.link)
		self.mainToolBar = aui.AuiToolBar(self.frame, wx.ID_ANY, wx.DefaultPosition, agwStyle=aui.AUI_TB_PLAIN_BACKGROUND)
		self.mainToolBar.SetToolBitmapSize(wx.Size(16, 16))
		self.AddTool(self.mainToolBar, 'new', 'img/new.png', _("New"), _("Create a new song"))
		self.AddTool(self.mainToolBar, 'open', 'img/open.png', _("Open"), _("Open an existing song"))
		self.AddTool(self.mainToolBar, 'save', 'img/save.png', _("Save"), _("Save song with the current filename"))
		self.mainToolBar.AddSeparator()
		self.undoTool = self.AddTool(self.mainToolBar, 'undo', 'img/undo.png', _("Undo"), _("Undo last edit"))
		self.redoTool = self.AddTool(self.mainToolBar, 'redo',
			'img/redo.png', _("Redo"), _("Redo previously undone edit"))
		self.redoTool = wx.xrc.XRCID('redo')
		self.mainToolBar.AddSeparator()
		self.cutTool = self.AddTool(self.mainToolBar, 'cut', 'img/cut.png', _("Cut"),
																_("Move selected text in the clipboard"))
		self.copyTool = self.AddTool(self.mainToolBar, 'copy', 'img/copy.png', _("Copy"),
																 _("Copy selected text in the clipboard"))
		self.copyOnlyTextTool = wx.xrc.XRCID('copyOnlyText')
		if platform.system() == 'Windows':
			self.AddTool(self.mainToolBar, 'copyAsImage', 'img/copyAsImage2.png', _("Copy as Image"),
									 _("Copy the whole FORMATTED song (or selected verses) to the clipboard"))
		self.pasteTool = self.AddTool(self.mainToolBar, 'paste', 'img/paste.png', _("Paste"),
																	_("Read text from the clipboard and place it at the cursor position"))
		self.pasteChordsTool = self.AddTool(self.mainToolBar, 'pasteChords', 'img/pasteChords.png', _("PasteChords"),
																				_("Integrate chords of copied text into current selection"))
		self.mainToolBar.Realize()
		self.mainToolBarPane = self.AddPane(self.mainToolBar, aui.AuiPaneInfo().ToolbarPane().Top().Row(1).Position(1),
																				_('Standard'), 'standard')
		self.formatToolBar = aui.AuiToolBar(self.frame, wx.ID_ANY, agwStyle=aui.AUI_TB_PLAIN_BACKGROUND)
		self.formatToolBar.SetExtraStyle(aui.AUI_TB_PLAIN_BACKGROUND)
		self.fontChooser = FontComboBox(self.formatToolBar, -1, self.pref.format.face)
		self.formatToolBar.AddControl(self.fontChooser)
		self.frame.Bind(wx.EVT_COMBOBOX, self.OnFontSelected, self.fontChooser)
		wx.UpdateUIEvent.SetUpdateInterval(500)
		self.frame.Bind(wx.EVT_UPDATE_UI, self.OnIdle, self.frame)
		self.frame.Bind(wx.EVT_TEXT_CUT, self.OnTextCutCopy, self.text)
		self.frame.Bind(wx.EVT_TEXT_COPY, self.OnTextCutCopy, self.text)
		self.fontChooser.Bind(wx.EVT_TEXT_ENTER, self.OnFontSelected, self.fontChooser)
		self.fontChooser.Bind(wx.EVT_KILL_FOCUS, self.OnFontSelected, self.fontChooser)
		self.AddTool(self.formatToolBar, 'title', 'img/title.png', _("Insert title"),
								 _("Insert a command to display song title"))
		self.AddTool(self.formatToolBar, 'chord', 'img/chord.png', _("Insert chord"),
								 _("Insert square brackets that will host a chord"))
		self.AddTool(self.formatToolBar, 'chorus', 'img/chorus.png', _("Insert chorus"),
								 _("Insert a couple of commands that will contain chorus"))
		self.AddTool(
			self.formatToolBar,
			'verseWithCustomLabelOrWithoutLabel',
			'img/verse.png',
			_("Insert verse with custom label or without label"),
			_("Insert a commands that will display a verse with a custom label"),
		)
		labelVersesTool = self.formatToolBar.AddToggleTool(  # AddToggleTool (agw) or AddTool
			wx.xrc.XRCID('labelVerses'),
			wx.Bitmap(wx.Image(glb.AddPath("img/labelVerses.png"))),
			wx.NullBitmap,
			True,
			None,
			_("Show verse labels"),
			_("Show or hide verse and chorus labels"),
		)
		self.labelVersesToolId = labelVersesTool.GetId()
		showChordsIcon = wx.StaticBitmap(self.formatToolBar, -1, wx.Bitmap(wx.Image(glb.AddPath('img/showChords.png'))))
		self.formatToolBar.AddControl(showChordsIcon)
		self.showChordsChooser = wx.Slider(self.formatToolBar, -1, 0, 0, 2, wx.DefaultPosition, (100, -1),
																			 wx.SL_AUTOTICKS | wx.SL_HORIZONTAL)
		tt1 = wx.ToolTip(_("Hide or show chords in formatted song"))
		tt2 = wx.ToolTip(_("Hide or show chords in formatted song"))
		self.showChordsChooser.SetToolTip(tt1)
		showChordsIcon.SetToolTip(tt2)
		self.frame.Bind(wx.EVT_SCROLL, self.OnFontSelected, self.showChordsChooser)
		self.formatToolBar.AddControl(
			self.showChordsChooser,
			"pippo"
		)
		self.formatToolBar.Realize()
		self.formatToolBarPane = self.AddPane(self.formatToolBar, aui.AuiPaneInfo().ToolbarPane().Top().Row(1).Position(2),
																					_('Format'), 'format')
		self.BindMyMenu()
		self.frame.Bind(EVT_TEXT_CHANGED, self.OnTextChanged)
		self.exportMenuId = xrc.XRCID('export')
		self.exportToClipboardAsAVectorImage = xrc.XRCID('exportToClipboardAsAVectorImage')
		self.exportAsEmfMenuId = xrc.XRCID('exportAsEmf')
		self.cutMenuId = xrc.XRCID('cut')
		self.copyMenuId = xrc.XRCID('copy')
		self.copyAsImageMenuId = xrc.XRCID('copyAsImage')
		self.pasteMenuId = xrc.XRCID('paste')
		self.pasteChordsMenuId = xrc.XRCID('pasteChords')
		self.removeChordsMenuId = xrc.XRCID('removeChords')
		self.labelVersesMenuId = xrc.XRCID('labelVerses')
		self.noChordsMenuId = xrc.XRCID('noChords')
		self.oneVerseForEachChordPatternMenuId = xrc.XRCID('oneVerseForEachChordPattern')
		self.wholeSongMenuId = xrc.XRCID('wholeSong')
		if platform.system() != 'Windows':
			self.menuBar.GetMenu(0).FindItemById(self.exportMenuId).GetSubMenu().Delete(self.exportToClipboardAsAVectorImage)
			self.menuBar.GetMenu(1).Delete(self.copyAsImageMenuId)
			self.menuBar.GetMenu(0).FindItemById(self.exportMenuId).GetSubMenu().Delete(self.exportAsEmfMenuId)
		self.findReplaceDialog = None
		self.CheckLabelVerses()
		self.SetFont()
		self.text.SetFont(self.pref.editorFace, self.pref.editorSize)
		self.FinalizePaneInitialization()
		# Reassign caption value to override caption saved in preferences (it could be another language)
		self._mgr.GetPane('preview').caption = _('Preview')
		self._mgr.GetPane('standard').caption = _('Standard')
		self._mgr.GetPane('format').caption = _('Format')
		if 'firstTimeEasyKey' in self.pref.notices:
			msg = _(
				"You are not a skilled guitarist? Songpress can help you: when you open a song, it can detect if chords are difficult. If this is the case, Songpress will alert you, and offer to transpose your song to the easiest key, automatically.\n\nDo you want to turn this option on?")
			d = wx.MessageDialog(self.frame, msg, _("Songpress"), wx.YES_NO | wx.ICON_QUESTION)
			if d.ShowModal() == wx.ID_YES:
				self.pref.autoAdjustEasyKey = True
				msg = _(
					"Please take a minute to set up your skill as a guitarist. For each group of chords, tell Songpress how much you like them.")
				d = wx.MessageDialog(self.frame, msg, _("Songpress"), wx.OK)
				d.ShowModal()
				f = MyPreferencesDialog(self.frame, self.pref, easyChords)
				f.notebook.SetSelection(1)
				if f.ShowModal() == wx.ID_OK:
					self.text.SetFont(self.pref.editorFace, int(self.pref.editorSize))
					self.SetDefaultExtension(self.pref.defaultExtension)
		MyUpdateDialog.check_and_update(self.frame, self.pref)

	def BindMyMenu(self):
		"""Bind a menu item, by xrc name, to a handler"""

		def Bind(handler, xrcname):
			self.Bind(wx.EVT_MENU, handler, xrcname)

		Bind(self.OnCopyAsImage, 'exportToClipboardAsAVectorImage')
		Bind(self.OnExportAsSvg, 'exportAsSvg')
		Bind(self.OnExportAsEmf, 'exportAsEmf')
		Bind(self.OnExportAsPng, 'exportAsPng')
		Bind(self.OnExportAsHtml, 'exportAsHtml')
		Bind(self.OnExportAsTab, 'exportAsTab')
		Bind(self.OnExportAsPptx, 'exportAsPptx')
		Bind(self.OnUndo, 'undo')
		Bind(self.OnRedo, 'redo')
		Bind(self.OnCut, 'cut')
		Bind(self.OnCopy, 'copy')
		Bind(self.OnCopyAsImage, 'copyAsImage')
		Bind(self.OnCopyOnlyText, 'copyOnlyText')
		Bind(self.OnPaste, 'paste')
		Bind(self.OnPasteChords, 'pasteChords')
		Bind(self.OnFind, 'find')
		Bind(self.OnFindNext, 'findNext')
		Bind(self.OnFindPrevious, 'findPrevious')
		Bind(self.OnReplace, 'replace')
		Bind(self.OnSelectNextChord, 'selectNextChord')
		Bind(self.OnSelectPreviousChord, 'selectPreviousChord')
		Bind(self.OnMoveChordRight, 'moveChordRight')
		Bind(self.OnMoveChordLeft, 'moveChordLeft')
		Bind(self.OnRemoveChords, 'removeChords')
		Bind(self.OnIntegrateChords, 'integrateChords')
		Bind(self.OnTitle, 'title')
		Bind(self.OnSubtitle, 'subtitle')
		Bind(self.OnChord, 'chord')
		Bind(self.OnChorus, 'chorus')
		Bind(self.OnVerse, 'verseWithCustomLabelOrWithoutLabel')
		Bind(self.OnComment, 'comment')
		Bind(self.OnFormatFont, 'songFont')
		Bind(self.OnTextFont, 'textFont')
		Bind(self.OnChordFont, 'chordFont')
		Bind(self.OnLabelVerses, 'labelVerses')
		Bind(self.OnChorusLabel, 'chorusLabel')
		Bind(self.OnNoChords, 'noChords')
		Bind(self.OnOneVerseForEachChordPattern, 'oneVerseForEachChordPattern')
		Bind(self.OnWholeSong, 'wholeSong')
		Bind(self.OnTranspose, 'transpose')
		Bind(self.OnSimplifyChords, 'simplifyChords')
		Bind(self.OnChangeChordNotation, 'changeChordNotation')
		Bind(self.OnNormalizeChords, 'cleanupChords')
		Bind(self.OnConvertTabToChordpro, 'convertTabToChordpro')
		Bind(self.OnRemoveSpuriousBlankLines, 'removeSpuriousBlankLines')
		Bind(self.OnOptions, 'options')
		Bind(self.OnGuide, 'guide')
		Bind(self.OnNewsAndUpdates, 'newsAndUpdates')
		Bind(self.OnDonate, 'donate')

	def AddTool(self, toolbar, resource_string, icon_path, label, help):
		tool = wx.xrc.XRCID(resource_string)
		toolbar.AddTool(
			tool,
			label,
			wx.Bitmap(wx.Image(glb.AddPath(icon_path))),
			wx.NullBitmap,
			wx.ITEM_NORMAL,
			label,
			help,
			None
		)
		return tool

	def New(self):
		self.text.AutoChangeMode(True)
		self.text.New()
		self.text.AutoChangeMode(False)
		self.UpdateEverything()

	def Open(self):
		self.text.AutoChangeMode(True)
		self.text.Open()
		self.text.AutoChangeMode(False)
		self.UpdateEverything()
		self.AutoAdjust(0, self.text.GetLength())

	def Save(self):
		self.text.Save()
		self.UpdateEverything()

	def SavePreferences(self):
		self.pref.Save()

	def UpdateUndoRedo(self):
		self.mainToolBar.EnableTool(self.undoTool, self.text.CanUndo())
		self.mainToolBar.EnableTool(self.redoTool, self.text.CanRedo())

	def UpdateCutCopyPaste(self):
		s, e = self.text.GetSelection()
		self.mainToolBar.EnableTool(self.cutTool, s != e)
		self.menuBar.Enable(self.cutMenuId, s != e)
		self.mainToolBar.EnableTool(self.copyTool, s != e)
		self.menuBar.Enable(self.copyOnlyTextTool, s != e)
		self.menuBar.Enable(self.copyMenuId, s != e)
		if platform.system() == 'Windows':
			cp = self.text.CanPaste()
		else:
			# Workaround for weird error in wxGTK
			cp = True
		self.mainToolBar.EnableTool(self.pasteTool, cp)
		self.menuBar.Enable(self.pasteMenuId, cp)
		self.mainToolBar.EnableTool(self.pasteChordsTool, cp)
		self.menuBar.Enable(self.pasteChordsMenuId, cp)
		self.menuBar.Enable(self.removeChordsMenuId, s != e)

	def UpdateEverything(self):
		self.UpdateUndoRedo()
		self.UpdateCutCopyPaste()

	def TextUpdated(self):
		self.previewCanvas.Refresh(self.text.GetText())

	# self.UpdateEverything()

	def DrawOnDC(self, dc):
		decorator = self.pref.decorator if self.pref.labelVerses else SongDecorator()
		r = Renderer(self.pref.format, decorator, self.pref.notations)
		start, end = self.text.GetSelection()
		if start == end:
			w, h = r.Render(self.text.GetText(), dc)
		else:
			w, h = r.Render(self.text.GetText(), dc, self.text.LineFromPosition(start), self.text.LineFromPosition(end))
		return w, h

	def AskExportFileName(self, type, ext):
		"""Ask the filename (without saving); return None if user cancels, the file name ow"""
		leave = False;
		consensus = False;
		while not leave:
			dlg = wx.FileDialog(
				self.frame,
				"Choose a name for the output file",
				"",
				os.path.splitext(self.document)[0],
				"%s files (*.%s)|*.%s|All files (*.*)|*.*" % (type, ext, ext),
				wx.FD_SAVE
			)

			if dlg.ShowModal() == wx.ID_OK:

				fn = dlg.GetPath()
				if os.path.isfile(fn):
					msg = "File \"%s\" already exists. Do you want to overwrite it?" % (fn,)
					d = wx.MessageDialog(
						self.frame,
						msg,
						self.appLongName,
						wx.YES_NO | wx.CANCEL | wx.ICON_QUESTION
					)
					res = d.ShowModal()
					if res == wx.ID_CANCEL:
						leave = True
						consensus = False
					elif res == wx.ID_NO:
						leave = False
						consensus = False
					else:  # wxID_YES
						leave = True
						consensus = True
				else:
					leave = True
					consensus = True

			else:
				leave = True
				consensus = False

		if consensus:
			return fn
		else:
			return None

	def OnExportAsPng(self, evt):
		n = self.AskExportFileName(_("PNG image"), "png")
		if n is not None:
			dc = wx.MemoryDC(wx.EmptyBitmap(1, 1))
			scale = 1
			w, h = self.DrawOnDC(dc)
			b = wx.EmptyBitmap(int(w * scale), int(h * scale))
			dc = wx.MemoryDC(b)
			dc.SetUserScale(scale, scale)
			dc.SetBackground(wx.WHITE_BRUSH);
			dc.Clear();
			self.DrawOnDC(dc)
			i = wx.ImageFromBitmap(b)
			i.SaveFile(n, wx.BITMAP_TYPE_PNG)

	def OnExportAsHtml(self, evt):
		n = self.AskExportFileName(_("HTML file"), "html")
		if n is not None:
			h = HtmlExporter(self.pref.format)
			r = Renderer(self.pref.format, h, self.pref.notations)
			start, end = self.text.GetSelection()
			if start == end:
				r.Render(self.text.GetText(), None)
			else:
				r.Render(self.text.GetText(), None, self.text.LineFromPosition(start), self.text.LineFromPosition(end))
			with open(n, "w", encoding='utf-8') as f:
				f.write(h.getHtml())

	def OnExportAsTab(self, evt):
		n = self.AskExportFileName(_("TAB file"), "tab")
		if n is not None:
			t = TabExporter(self.pref.format)
			r = Renderer(self.pref.format, t, self.pref.notations)
			start, end = self.text.GetSelection()
			if start == end:
				r.Render(self.text.GetText(), None)
			else:
				r.Render(self.text.GetText(), None, self.text.LineFromPosition(start), self.text.LineFromPosition(end))
			with open(n, "w", encoding='utf-8') as f:
				f.write(t.getTab())

	def OnExportAsSvg(self, evt):
		n = self.AskExportFileName(_("SVG image"), "svg")
		if n is not None:
			dc = wx.MemoryDC(wx.Bitmap(1, 1))
			w, h = self.DrawOnDC(dc)
			dc = wx.SVGFileDC(n, int(w), int(h))
			self.DrawOnDC(dc)

	def OnExportAsEmf(self, evt):
		n = self.AskExportFileName(_("Enhanced Metafile"), "emf")
		if n is not None:
			dc = wx.msw.MetafileDC(n)
			self.DrawOnDC(dc)
			dc.Close()

	def OnExportAsEps(self, evt):
		n = self.AskExportFileName(_("EPS image"), "eps")
		if n is not None:
			pd = wx.PrintData()
			pd.SetPaperId(wx.PAPER_NONE)
			pd.SetPrintMode(wx.PRINT_MODE_FILE)
			pd.SetFilename(n)
			dc = wx.PostScriptDC(pd)
			dc.StartDoc(_("Exporting image as EPS..."))
			self.DrawOnDC(dc)
			dc.EndDoc()


	def OnExportAsPptx(self, evt):
		try:
			import songimpress
		except ImportError:
			msg = _("Please install the python-pptx module to use this feature")
			d = wx.MessageDialog(self.frame, msg, "Songpress", wx.OK | wx.ICON_ERROR)
			d.ShowModal()
			return
		text = replaceTitles(self.text.GetTextOrSelection(), '---')
		text = removeChordPro(text).strip()
		if text != '':
			template_rel = os.path.join('templates', 'slides')
			template_paths = [f for f in glb.ListLocalGlobalDir(template_rel) if f[-5:].upper() == '.PPTX']
			template_names = [os.path.split(f)[1][:-5] for f in template_paths]
			mld = MyListDialog(
				self.frame,
				_("Please select a template for your PowerPoint presentation:"),
				_("Export as PowerPoint"),
				template_names,
			)
			if mld.ShowModal() == wx.ID_OK:
				output_file = self.AskExportFileName(_("PPTX presentation"), "pptx")
				if output_file is not None:
					i = mld.GetSelectedIndex()
					songimpress.to_presentation(text.splitlines(), output_file, template_paths[i])

	def OnUpdateUI(self, evt):
		self.UpdateEverything()
		evt.Skip()

	def OnUndo(self, evt):
		if self.text.CanUndo():
			self.text.Undo()
			self.UpdateUndoRedo()

	def OnRedo(self, evt):
		if self.text.CanRedo():
			self.text.Redo()
			self.UpdateUndoRedo()

	def OnCut(self, evt):
		self.text.Cut()

	def OnTextCutCopy(self, evt):
		self.UpdateCutCopyPaste()
		evt.Skip()

	def OnCopy(self, evt):
		self.text.Copy()

	def OnCopyOnlyText(self, evt):
		self.text.CopyOnlyText()

	def OnCopyAsImage(self, evt):
		dc = wx.msw.MetafileDC()
		self.DrawOnDC(dc)
		m = dc.Close()
		m.SetClipboard(dc.MaxX(), dc.MaxY())

	def OnPaste(self, evt):
		self.text.Paste()

	def OnPasteChords(self, evt):
		self.text.PasteChords()

	def OnFind(self, evt):
		self.findReplaceDialog = SongpressFindReplaceDialog(self)

	def OnFindNext(self, evt):
		if self.findReplaceDialog != None:
			self.findReplaceDialog.down = True
			self.findReplaceDialog.FindNext()

	def OnFindPrevious(self, evt):
		if self.findReplaceDialog != None:
			self.findReplaceDialog.down = False
			self.findReplaceDialog.FindNext()

	def OnReplace(self, evt):
		self.findReplaceDialog = SongpressFindReplaceDialog(self, True)

	def OnSelectNextChord(self, evt):
		self.text.SelectNextChord()

	def OnSelectPreviousChord(self, evt):
		self.text.SelectPreviousChord()

	def OnMoveChordRight(self, evt):
		r = self.text.GetChordUnderCursor()
		if r is not None:
			n = self.text.GetLength()
			s, e, c = r
			if e < n:
				self.text.BeginUndoAction()
				e1 = self.text.PositionAfter(e)
				self.text.SetSelection(e, e1)
				l = self.text.GetTextRange(e, e1)
				self.text.ReplaceSelection('')
				self.text.SetSelection(s, s)
				self.text.ReplaceSelection(l)
				s2 = self.text.PositionAfter(self.text.PositionAfter(s))
				self.text.SetSelection(s2, s2)
				self.text.EndUndoAction()

	def OnMoveChordLeft(self, evt):
		r = self.text.GetChordUnderCursor()
		if r is not None:
			s, e, c = r
			if s > 0:
				self.text.BeginUndoAction()
				s1 = self.text.PositionBefore(s)
				e1 = self.text.PositionBefore(e)
				l = self.text.GetTextRange(s1, s)
				self.text.SetSelection(e, e)
				self.text.ReplaceSelection(l)
				self.text.SetSelection(s1, s)
				self.text.ReplaceSelection('')
				s = self.text.PositionAfter(s1)
				self.text.SetSelection(s, s)
				self.text.EndUndoAction()

	def OnRemoveChords(self, evt):
		self.text.RemoveChordsInSelection()

	def OnIntegrateChords(self, evt):
		ln = self.text.GetCurrentLine()
		if ln < self.text.GetLineCount() - 1:
			chords = self.text.GetLine(ln).strip("\r\n")
			text = self.text.GetLine(ln + 1).strip("\r\n")
			chordpro = integrateChords(chords, text)
			self.text.SetSelectionStart(self.text.PositionFromLine(ln))
			self.text.SetSelectionEnd(self.text.GetLineEndPosition(ln + 1))
			self.text.ReplaceSelection(chordpro)

	def OnFontSelected(self, evt):
		font = self.fontChooser.GetValue()
		showChords = self.showChordsChooser.GetValue()
		self.pref.SetFont(font, showChords)
		self.SetFont(True)
		evt.Skip()

	def OnGuide(self, evt):
		if platform.system() == 'Windows':
			helpfile = os.path.join("../help", "songpress-%s.chm" % (i18n.getLang(),))
			subprocess.Popen("hh " + glb.AddPath(helpfile))
		else:
			wx.LaunchDefaultBrowser(_("http://www.skeed.it/songpress-manual"))

	def OnIdle(self, evt):
		try:
			cp = self.text.CanPaste()
			self.mainToolBar.EnableTool(self.pasteTool, cp)
			self.menuBar.Enable(self.pasteMenuId, cp)
			self.mainToolBar.EnableTool(self.pasteChordsTool, cp)
			self.menuBar.Enable(self.pasteChordsMenuId, cp)
		except wx._core.PyDeadObjectError:
			# When frame is closed, this method may still be executed, generating an exception
			# because UI elements have been destroyed. Simply ignore it.
			pass
		evt.Skip()

	def OnNewsAndUpdates(self, evt):
		MyUpdateDialog.check_and_update(self.frame, self.pref, True)

	def OnDonate(self, evt):
		wx.LaunchDefaultBrowser(_("http://www.skeed.it/songpress#donate"))

	def OnFormatFont(self, evt):
		f = FontFaceDialog(self.frame, wx.ID_ANY, _("Songpress"), self.pref.format, self.pref.decorator,
											 self.pref.decoratorFormat)
		if f.ShowModal() == wx.ID_OK:
			self.pref.SetFont(f.GetValue())
			self.SetFont()

	def OnTextFont(self, evt):
		data = wx.FontData()
		data.SetInitialFont(self.pref.format.wxFont)
		data.SetColour(self.pref.format.color)

		dialog = wx.FontDialog(self.frame, data)
		if dialog.ShowModal() == wx.ID_OK:
			retData = dialog.GetFontData()
			font = retData.GetChosenFont()
			face = font.GetFaceName()
			size = font.GetPointSize()
			color = retData.GetColour().GetAsString(wx.C2S_HTML_SYNTAX)
			s = f"{{textfont:{face}}}{{textsize:{size}}}{{textcolour:{color}}}|"
			s += "{textfont}{textsize}{textcolour}"
			self.InsertWithCaret(s)

	def OnChordFont(self, evt):
		data = wx.FontData()
		data.SetInitialFont(self.pref.format.wxFont)
		data.SetColour(self.pref.format.color)

		dialog = wx.FontDialog(self.frame, data)
		if dialog.ShowModal() == wx.ID_OK:
			retData = dialog.GetFontData()
			font = retData.GetChosenFont()
			face = font.GetFaceName()
			size = font.GetPointSize()
			color = retData.GetColour().GetAsString(wx.C2S_HTML_SYNTAX)
			s = f"{{chordfont:{face}}}{{chordsize:{size}}}{{chordcolour:{color}}}|"
			s += "{chordfont}{chordsize}{chordcolour}"
			self.InsertWithCaret(s)

	def OnTranspose(self, evt):
		t = MyTransposeDialog(self.frame, self.pref.notations, self.text.GetTextOrSelection())
		if t.ShowModal() == wx.ID_OK:
			self.text.ReplaceTextOrSelection(t.GetTransposed())

	def OnSimplifyChords(self, evt):
		self.text.AutoChangeMode(True)
		t = self.text.GetTextOrSelection()
		notation = autodetectNotation(t, self.pref.notations)
		count, c, dc, e, de = findEasiestKey(t, self.pref.GetEasyChords(), notation)
		title = _("Simplify chords")
		if count > 0 and dc != de:
			msg = _("The key of your song, %s, is not the easiest to play (difficulty: %.1f/5.0).\n") % (c, 5 * dc)
			msg += _("Do you want to transpose the key %s, which is the easiest one (difficulty: %.1f/5.0)?") % (e, 5 * de)
			d = wx.MessageDialog(self.frame, msg, title, wx.YES_NO | wx.ICON_QUESTION)
			if d.ShowModal() == wx.ID_YES:
				t = transposeChordPro(translateChord(c, notation), translateChord(e, notation), t, notation)
				self.text.ReplaceTextOrSelection(t)
		else:
			if count > 0:
				msg = _("The key of your song, %s, is already the easiest to play (difficulty: %.1f/5.0).\n") % (c, 5 * dc)
			else:
				msg = _("Your song or current selection does not contain any chords.")
			d = wx.MessageDialog(self.frame, msg, title, wx.OK | wx.ICON_INFORMATION)
			d.ShowModal()
		self.text.AutoChangeMode(False)

	def OnChangeChordNotation(self, evt):
		t = MyNotationDialog(self.frame, self.pref.notations, self.text.GetTextOrSelection())
		if t.ShowModal() == wx.ID_OK:
			self.text.ReplaceTextOrSelection(t.ChangeChordNotation())

	def OnNormalizeChords(self, evt):
		t = MyNormalizeDialog(self.frame, self.pref.notations, self.text.GetTextOrSelection())
		if t.ShowModal() == wx.ID_OK:
			self.text.ReplaceTextOrSelection(t.NormalizeChords())

	def OnConvertTabToChordpro(self, evt):
		t = self.text.GetTextOrSelection()
		n = testTabFormat(t, self.pref.notations)
		if n is not None:
			self.text.ReplaceTextOrSelection(tab2ChordPro(t, n))

	def OnRemoveSpuriousBlankLines(self, evt):
		self.text.ReplaceTextOrSelection(removeSpuriousLines(self.text.GetTextOrSelection()))

	def OnOptions(self, evt):
		f = MyPreferencesDialog(self.frame, self.pref, easyChords)
		if f.ShowModal() == wx.ID_OK:
			self.text.SetFont(self.pref.editorFace, int(self.pref.editorSize))
			self.SetDefaultExtension(self.pref.defaultExtension)

	def StripSelection(self):
		"""
		Update selection, moving blank characters out of it
		"""
		s, e = self.text.GetSelection()
		mod = False
		while e > s and self.text.GetTextRange((ep := self.text.PositionBefore(e)), e).strip() == '':
			e = ep
			mod = True
		while s < e and self.text.GetTextRange(s, (sa := self.text.PositionAfter(s))).strip() == '':
			s = sa
			mod = True
		if mod:
			self.text.SetSelection(s, e)

	def InsertWithCaret(self, st):
		self.StripSelection()
		s, e = self.text.GetSelection()
		c = st.find('|')
		if c != -1:
			sel_text = self.text.GetSelectedText()
			self.text.ReplaceSelection(st[:c] + sel_text + st[c + 1:])
			self.text.SetSelection(s + c, e + c)
		else:
			self.text.ReplaceSelection(st)
			self.text.SetSelection(s + len(st), s + len(st))

	def OnTitle(self, evt):
		self.InsertWithCaret("{title:|}")

	def OnSubtitle(self, evt):
		self.InsertWithCaret("{subtitle:|}")

	def OnChord(self, evt):
		self.InsertWithCaret("[|]")

	def OnVerse(self, evt):
		label = wx.GetTextFromUser(
			_("Insert a label for verse, or press Cancel if you want to omit label."),
			_("Verse label"),
			"",
			self.frame,
		)
		self.InsertWithCaret("{Verse:%s}|" % label)

	def OnChorus(self, evt):
		default = self.pref.decoratorFormat.GetChorusLabel()
		label = wx.GetTextFromUser(
			_("Insert a label for chorus, or press Cancel if you want to omit label."),
			_("Chorus label"),
			default,
			self.frame,
		)
		if label == default:
			self.InsertWithCaret("{soc}\n|\n{eoc}\n")
		else:
			self.InsertWithCaret("{soc:%s}\n|\n{eoc}\n" % label)

	def OnComment(self, evt):
		self.InsertWithCaret("{c:|}")

	def OnLabelVerses(self, evt):
		self.pref.labelVerses = not self.pref.labelVerses
		self.CheckLabelVerses()

	def OnChorusLabel(self, evt):
		c = self.pref.decoratorFormat.GetChorusLabel()
		msg = _("Please enter a label for chorus")
		d = wx.TextEntryDialog(self.frame, msg, _("Songpress"), c)
		if d.ShowModal() == wx.ID_OK:
			c = d.GetValue()
			self.pref.SetChorusLabel(c)
			self.previewCanvas.Refresh(self.text.GetText())

	def OnNoChords(self, evt):
		self.pref.format.showChords = 0
		self.SetFont(True)

	def OnOneVerseForEachChordPattern(self, evt):
		self.pref.format.showChords = 1
		self.SetFont(True)

	def OnWholeSong(self, evt):
		self.pref.format.showChords = 2
		self.SetFont(True)

	def OnTextChanged(self, evt):
		self.AutoAdjust(evt.lastPos, evt.currentPos)

	def AutoAdjust(self, lastPos, currentPos):
		self.text.AutoChangeMode(True)
		t = self.text.GetTextRange(lastPos, currentPos)
		if self.pref.autoAdjustSpuriousLines:
			if testSpuriousLines(t):
				msg = _("It looks like there are spurious blank lines in the song.\n")
				msg += _("Do you want to try to remove them automatically?")
				title = _("Remove spurious blank lines")
				d = wx.MessageDialog(self.frame, msg, title, wx.YES_NO | wx.ICON_QUESTION)
				if d.ShowModal() == wx.ID_YES:
					self.text.SetSelection(lastPos, currentPos)
					t = removeSpuriousLines(t)
					self.text.ReplaceSelection(t)
					currentPos = self.text.GetCurrentPos()
		if self.pref.autoAdjustTab2Chordpro:
			n = testTabFormat(t, self.pref.notations)
			if n is not None:
				msg = _("It looks like your song is in tab format (i.e., chords are above the text).\n")
				msg += _("Do you want to convert it to ChordPro automatically?")
				title = _("Convert to ChordPro")
				d = wx.MessageDialog(self.frame, msg, title, wx.YES_NO | wx.ICON_QUESTION)
				if d.ShowModal() == wx.ID_YES:
					self.text.SetSelection(lastPos, currentPos)
					t = tab2ChordPro(t, n)
					self.text.ReplaceSelection(t)
		if self.pref.autoAdjustEasyKey:
			notation = autodetectNotation(t, self.pref.notations)
			count, c, dc, e, de = findEasiestKey(t, self.pref.GetEasyChords(), notation)
			if count > 10 and dc != de:
				msg = _("The key of your song, %s, is not the easiest to play (difficulty: %.1f/5.0).\n") % (c, 5 * dc)
				msg += _("Do you want to transpose the key %s, which is the easiest one (difficulty: %.1f/5.0)?") % (e, 5 * de)
				title = _("Simplify chords")
				d = wx.MessageDialog(self.frame, msg, title, wx.YES_NO | wx.ICON_QUESTION)
				if d.ShowModal() == wx.ID_YES:
					self.text.SetSelection(lastPos, currentPos)
					t = transposeChordPro(translateChord(c, notation), translateChord(e, notation), t, notation)
					self.text.ReplaceSelection(t)
		self.text.AutoChangeMode(False)

	def SetFont(self, updateFontChooser=True):
		try:
			if updateFontChooser:
				self.fontChooser.SetValue(self.pref.format.face)
				self.showChordsChooser.SetValue(self.pref.format.showChords)
				ids = [self.noChordsMenuId, self.oneVerseForEachChordPatternMenuId, self.wholeSongMenuId]
				self.menuBar.Check(ids[self.pref.format.showChords], True)

			self.previewCanvas.Refresh(self.text.GetText())
		except wx._core.PyDeadObjectError:
			# When frame is closed, this method may still be executed, generating an exception
			# because UI elements have been destroyed. Simply ignore it.
			pass

	def CheckLabelVerses(self):
		self.formatToolBar.ToggleTool(self.labelVersesToolId, self.pref.labelVerses)
		self.formatToolBar.Refresh()
		self.menuBar.Check(self.labelVersesMenuId, self.pref.labelVerses)
		if self.pref.labelVerses:
			self.previewCanvas.SetDecorator(self.pref.decorator)
		else:
			self.previewCanvas.SetDecorator(SongDecorator())
		self.previewCanvas.Refresh(self.text.GetText())


